# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

# Python 3 imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# For version stuff...
import sys

# For timing
import time

# For exception reporting/logging
import traceback

# Minimum python version
VERSION_PYTHON_MINIMUM =                                (2, 7)

def main_wrapper():
    """Wraps main() in some handy niceness."""

    # Check python version first.
    if sys.version_info < VERSION_PYTHON_MINIMUM:
        print("Dupligänger requires Python {0} or later.  You are running {1}.".format(
            '.'.join([str(i) for i in VERSION_PYTHON_MINIMUM]),
            '.'.join([str(i) for i in sys.version_info[0:3]])))
        sys.exit(1)

    # Now you can safely do Dupligänger imports
    import dupliganger.dupliganger

    # Go
    dupliganger.dupliganger.main()


# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object
# py-lmdb uses bytes() for py3 and str() for py2.
# This package has different code for py2 and py3.
# And importing that future 'object' has a bug that screws up __slots__ in
# py2 (causes different behavior than in py3).

# Dupligänger imports

## Other imports

import pytest
import os
import shutil

#################
### Constants ###
#################

LAST_RUN_DIR            = 'last_run'
IN_DIR                  = 'in'
EXPECTED_OUT_DIR        = 'eout'
BASE_FILES_DIR          = 'dupliganger/test/files'
                          # e.g. 'dupliganger/test/files/last_run')
BASE_OUTFILE_DIR        = os.path.join(BASE_FILES_DIR, LAST_RUN_DIR)


#################
### Functions ###
#################

@pytest.fixture(scope="module")
def test_outdir(request):
    """Output directory for the test run."""

    # request.module == the name of the test module being run, e.g. test_remove_umi.py

    # Strip off the prepended 'test_'
    # type = module -> 'test_remove_umi' -> 'remove_umi'
    command_being_tested = request.module.__name__[5:]

    _test_outdir = os.path.join(BASE_OUTFILE_DIR, command_being_tested)

    # Remove any existing directory
    if _test_outdir[0] == '/':
        # safety check
        raise

    # Remove trace of any previous run
    shutil.rmtree(_test_outdir, True)

    # create the directory
    os.mkdir(_test_outdir)

    # run the tests
    yield _test_outdir

    # Remove the directory
    if not os.environ.get("DUPLIGANGER_KEEP_TEST_DIRS"):
        shutil.rmtree(_test_outdir, True)

def to_eout_filename(fout_filename, python_version=None):
    """Convert filename of output file to filename of expected output file.

    If user passes optional python_version, then the tests expect different
    outputs for python version 2 vs version 3 (at time of writing, the only
    reason for this is the difference in python 2/3 pseudo random number
    generator) and so the files are put in a subdir named either 'py2' or
    'py3'.

    Example:
        'files/last_run/remove_adapter/A_R1.rmadapt.fq'
        converts to
        'files/remove_adapter/eout/A_R1.rmadapt.fq'

    Example of passing python_version=3:
        'files/last_run/remove_adapter/A_R1.rmadapt.fq'
        converts to
        'files/remove_adapter/eout/py3/A_R1.rmadapt.fq'

    Args:
        python_version (str): Optional; user specified a python_version (see
            above).
    """
    # Check reasonable python_version
    assert python_version in (2, 3, None)

    elems = fout_filename.split('/')
    # -> ['files', 'last_run', 'remove_adapter', 'A_R1.rmadapt.fq']
    just_filename = elems[-1]
    elems.remove(LAST_RUN_DIR)
    # -> ['files', 'remove_adapter', 'A_R1.rmadapt.fq']
    elems.insert(-1, EXPECTED_OUT_DIR)
    # -> ['files', 'remove_adapter', 'eout', 'A_R1.rmadapt.fq']
    if python_version is not None:
        elems.insert(-1, "py{}".format(python_version))
        # -> ['files', 'remove_adapter', 'eout', 'py3', 'A_R1.rmadapt.fq']
    feout_filename = '/'.join(elems)
    # -> 'files/remove_adapter/eout/A_R1.rmadapt.fq'

    # Leave these print statements here, otherwise very difficult to debug tox
    # output when one fails (because we're using pytest fixtures, so without
    # these print statements, impossible to tell which file it failed on).
    print("actual   output filename = {}".format(fout_filename))
    print("expected output filename = {}".format(feout_filename))
    print("just     the    filename = {}".format(just_filename))

    return feout_filename

def to_eout_db_dump_filename(alignment_file, db_name):
    """Convert filename of output file to filename of expected output file.

    Args:
        alignment_file (str): Name of the alignment file.
        db_name (str): Name of the db
    Example:
        args ('files/build_read_db/in/11_first_test.pe.sam', 'read_group_db')
        converts to
        'files/build_read_db/eout/11_first_test.pe.read_group_db'
    """

    if alignment_file[-4:] in ('.sam', '.bam'):
        # Chop off .sam or .bam extension
        alignment_file = alignment_file[:-4]

    elems = alignment_file.split('/')
    # -> ['files', 'build_read_db', 'in', '11_first_test.pe']
    elems.remove(IN_DIR)
    # -> ['files', 'build_read_db', '11_first_test.pe']
    elems.insert(-1, EXPECTED_OUT_DIR)
    # -> ['files', 'build_read_db', 'eout', '11_first_test.pe']
    eout = '/'.join(elems)
    # -> 'files/build_read_db/eout/11_first_test.pe'
    return '.'.join((eout, db_name))
    # -> 'files/build_read_db/eout/11_first_test.pe.read_group_db'

def fix_paths(args, base_in, base_out, file_params):
    """Takes as input the docopt generated 'args' dict, and fixes all of the
    input/output paths.

    Example:
        args before = { ...
            '-o': None,
            '<in1.fastq>': 'A_R1.fq',
            '<in2.fastq>': 'A_R2.fq',
            '<input.fastq>': None,
        ... }
        args after = { ...
            '-o': 'dupliganger/test/files/last_run/remove_adapter',
            '<in1.fastq>': 'dupliganger/test/files/remove_adapter/in/A_R1.fq',
            '<in2.fastq>': 'dupliganger/test/files/remove_adapter/in/A_R2.fq',
            '<input.fastq>': None,
        ... }

    Args:
        args (dict): docopt generated args dictionary.
        base_in (str): Directory where the input files are stored.
            e.g. 'dupliganger/test/files/remove_adapter/in'
        base_out (str): Directory where the output files are placed.
            e.g. 'dupliganger/test/files/last_run/remove_adapter/'
        file_params ([str]): e.g. ['<input.fastq>' '<in1.fastq>', '<in2.fastq>']

    Returns:
        dict: The modified args dict.

    """
    # Fix output dir path
    args['-o'] = base_out

    # Fix input file paths
    for param in file_params:
        filename = args[param]
        if filename is not None:
            args[param] = os.path.join(base_in, filename)
    return args

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Build databases of aligned reads and their locations.

You probably don't want to run this command.

'build-read-and-loc-dbs' is a subcommand of the 'dedup' command.

You probably want to run the 'dedup' command instead.


Usage:
    dupliganger build-read-and-loc-dbs [options] <alignment-file>


Options:
    -h, --help
    -v, --verbose        Be verbose
    -o OUT_DIR           Place results in directory OUT_DIR.
    -k KIT, --kit KIT    The kit used [default: bioo].
    -t N, --threads N    Number of threads.
    --store STORE        Storage backend: 'lmdb' or 'memory' [default: lmdb]
    --debug-switch S     Fun debug switch [default: nothing].
    --debug-dump-rg-db   Dump the ReadGroup database (for debugging purposes only).
    --debug-dump-loc-db  Dump the Location database (for debugging purposes only).

"""

###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object
# py-lmdb uses bytes() for py3 and str() for py2.
# This package has different code for py2 and py3.
# And importing that future 'object' has a bug that screws up __slots__ in
# py2 (causes different behavior than in py3).

from builtins import ascii

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *
from dupliganger.common import (setup_report_db, sambamopen, args_to_out_dir,
        memory_info)
from dupliganger.db import (ParentDbDict, ParentDbLmdb, SimpleBucketDict,
        SimpleBucketLmdb, LocationBucketDb, SimpleObjectDbDict,
        SimpleObjectDbLmdb)
from dupliganger.sam import (Read, ReadGroup, to_location_key_with_5p_trimming)

# debug
from dupliganger.db import DebugMultipleLocationBucketDbs

## Other imports
from docopt import docopt

# For filename fixing
import os

# For debug timing
import time

# For some printing to stdout
import sys

#################
### Constants ###
#################


#################
### Functions ###
#################

def write_to_read_and_location_dbs(report_db, fin, parent_db, read_group_db,
        location_db, records_per_txn):
    """Read SAM file, populate ReadGroupDb and BucketDb..."""
    first_loop = True
    done = False
    time1 = None
    time2 = None
    time_in_write_db = 0
    read_group_id = 1
    curr_read_group = ReadGroup()

    while (not done):
        # Writing read_group_db (and optionally, location_db)...
        if first_loop:
            first_loop = False
        else:
            time2 = time.time()
            time_in_write_db += (time2 - time1)
            print("Commit at read_group_id = {}, time taken = {}, current mem MBs: {}".format(
                    read_group_id, time2 - time1, memory_info(True)))
        time1 = time.time()
        with parent_db.begin(True) as txn:
            # functions
            fp_put_read_group = read_group_db.fp_put(txn)
            # fp_put_read_name = read_name_db.fp_put(txn)
            fp_append_location_db = location_db.fp_append(txn)
            (done, read_group_id, curr_read_group) = (
                    write_to_read_and_location_dbs_txn(report_db, fin,
                        read_group_id, curr_read_group, records_per_txn,
                        fp_put_read_group, fp_append_location_db))

def write_to_read_and_location_dbs_txn(report_db, fin, read_group_id,
        curr_read_group, records_per_txn, fp_put_read_group,
        fp_append_location_db):
    """
    Write a 'records_per_txn' number of reads and optionally locations to DB.

    Args:
        read_group_id (int): The current (or starting) read_group_id.
        curr_read_group (ReadGroup): Any leftover read_groups from previous txn are
            passed along here.
        records_per_txn (int): The number of records
        fp_put_read_group (function): Function to add a read_group to the
            read_group database.
        fp_append_location_db (function): Function to append a read_group_id to the
            genomic location database.
    Returns:
        (bool, int, [str]): 3-tuple with elements a) done or not, b) current
            read_group_id, c) current unwritten/incomplete read_group.

    """
    record_count = 0
    total_count = 0
    read = None
    remaining_lines = True
    # Used to keep keys ordered.
    read_group_id_str = ascii(read_group_id).zfill(READ_GROUP_ID_DIGITS)

    if len(curr_read_group) == 0:
        # First txn
        read_group = ReadGroup()
        prev_read = None
    else:
        # Leftover read_group not written in last txn.
        read_group = curr_read_group
        prev_read = curr_read_group[-1]

    while remaining_lines and record_count < records_per_txn:
        while True: # Assumption: average len(read_group) << records_per_txn
            # Collect all reads with same RNAME
            total_count += 1
            line = fin.readline().rstrip()
            if not line:
                read = None
                remaining_lines = False
                break
            if line[0] == '@':
                # skip header lines
                continue
            read = Read(line)
            if prev_read and prev_read.qname != read.qname:
                # New read name found, write to store current read_group.
                break
            else:
                read_group.append(read)
                prev_read = read

        if prev_read:
            # Writing to store
            read_name = read_group[0].qname.split(DELIM_ANNO)[0]
            fp_put_read_group(read_group_id_str, read_group)
            fp_append_location_db(read_group_id_str, read_group)
            report_db[LOG_NUM_READ_GROUPS] += 1

        # Prep for next round
        read_group = ReadGroup([read]) if read else ReadGroup()

        prev_read = read
        record_count += 1
        read_group_id += 1
        read_group_id_str = ascii(read_group_id).zfill(READ_GROUP_ID_DIGITS)
        if not remaining_lines:
            break

    if not remaining_lines and len(read_group) > 0:
        # Get the very last read_group of the file
        fp_put_read_group(read_group_id_str, read_group)
        fp_append_location_db(read_group_id_str, read_group)
        report_db[LOG_NUM_READ_GROUPS] += 1

    done = not remaining_lines
    return (done, read_group_id, read_group)

def parse_args(args):
    """Parse the command line arguments."""
    debug_switch = args['--debug-switch']
    dump_rg_db = args['--debug-dump-rg-db']
    dump_loc_db = args['--debug-dump-loc-db']

    # Convert ~ to real path
    input_file = os.path.expanduser(args['<alignment-file>'])

    # Which kit?
    kit = args['--kit']
    if kit == KIT_BIOO:
        pass
    else:
        raise CannotContinueException("""Kit {} is not supported.""".format(kit))

    # Figure out which function to use to write to output file.
    num_threads = args['--threads']

    # Which store to use
    if args['--store'] not in (STORE_OPTION_LMDB, STORE_OPTION_MEMORY):
        raise CannotContinueException("""Store {} is not supported.""".format(args['--store']))
    store = args['--store']

    outdir = args_to_out_dir(args)

    return (kit, store, outdir, input_file, debug_switch, dump_rg_db,
            dump_loc_db)

def run(kit, store, outdir, input_file, debug_switch, dump_rg_db, dump_loc_db):
    """Start the run.

    Args:
        kit (str): kit...
        store (str): Which storage backend to use.
        outdir (str): Output directory for results
        input_dir (str): ...
    """

    if not os.path.isfile(input_file):
        raise CannotContinueException('Input file {} does not exist.'.format(
                input_file))

    report_db = setup_report_db()
    if store == STORE_OPTION_LMDB:
        db_file = os.path.join(outdir, os.path.split(input_file)[1] + '.sdd.db')
        parent_db = ParentDbLmdb(db_file, LMDB_MAX_DBS, LMDB_DB_SIZE)
        read_group_db = SimpleObjectDbLmdb('read_group', parent_db, parent_db.env, ReadGroup)
        # read_name_db = SimpleObjectDbLmdb('read_name', parent_db, parent_db.env, int)
        location_bucket_store = SimpleBucketLmdb('location_bucket_store',
                parent_db, parent_db.env, DELIM_BUCKET_LIST, str)
    elif store == STORE_OPTION_MEMORY:
        parent_db = ParentDbDict()
        read_group_db = SimpleObjectDbDict()
        # read_name_db = SimpleObjectDbDict()
        location_bucket_store = SimpleBucketDict()

    loc_db = LocationBucketDb(location_bucket_store, to_location_key_with_5p_trimming)

    with sambamopen(input_file) as fin:
        write_to_read_and_location_dbs(report_db, fin, parent_db,
                read_group_db, loc_db, RECORDS_PER_TXN)

    if dump_rg_db:
        sys.stderr.write(str(read_group_db))
    if dump_loc_db:
        sys.stderr.write(str(loc_db))

    return (parent_db, read_group_db, loc_db)


###############
### Classes ###
###############


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Annotate quality trimmed files with number of 5' and 3' NTs trimmed.

'annotate-qtrim' uses the 'trimlog' produced by Trimmomatic to annotate.

Usage:
    dupliganger annotate-qtrim [options] <input.fastq>
    dupliganger annotate-qtrim [options] <in1.fastq> <in2.fastq>

Note:
    Dupligänger supports (and autodetects) input FASTQ files that are gzipped.

Options:
    -h, --help
    -v, --verbose         Be verbose.
    -l TRIMLOG            Optionally specify trimlog.
    -o OUT_DIR            Place results in directory OUT_DIR.
    --compress            Compress (gzip) output files.
    -t N, --threads N     EXPERIMENTAL: If pigz is installed and --threads is
                          specified, output FASTQ files will be compressed with
                          pigz -p <n>; otherwise, they will be left
                          uncompressed (as it simply takes too long to compress
                          with just gzip) [default: 1].
"""

###############
### Imports ###
###############

# Python 3 imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from builtins import range

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *
from dupliganger.common import (pgopen, tmpf_start, tmpf_finish,
        is_gzipped, filename_in_to_out_fqgz, pe_log_filename, se_log_filename,
        args_to_out_dir)

## Other imports
from docopt import docopt

# For converting ~ to full path
import os

# For shell-like "which()"
try:
    from shutil import which
except ImportError:
    from whichcraft import which

# For function partials
import functools

# For inspecting gzipped files
import gzip


#################
### Constants ###
#################

#################
### Functions ###
#################

def create_annotated_files(in1, in2, trim_log, out1, out2, has_index):
    """Creates paired-end files with read names annotated with amount of 5'-trimming.

    Example:
        "ReadName-ACGT|GGCC" -> "ReadName-ACGT|GGCC;1^5"

    Args:
        in1 (file): Read1 fastq input filehandle.
        in2 (file): Read2 fastq input filehandle.
        trim_log (file): Trimmomatic trim log.
        out1 (file): Read1 annotated fastq output filehandle.
        out2 (file): Read2 annotated fastq output filehandle.
        has_index (bool): Whether or not these FASTQ files have the index
            string (e.g. 'read_id 1:N:0:ATCACGTT')
    """
    eof = False
    while True:
        name1, seq1, _, qual1 = \
                (in1.readline().rstrip() for i in range(4))
        name2, seq2, _, qual2 = \
                (in2.readline().rstrip() for i in range(4))

        if not name1 or not name2:
            break

        if has_index:
            name1, index_str1 = name1.split()
            name2, index_str2 = name2.split()

        # examples:
        # D00597:180:C7NMDANXX:6:1101:1184:35164-ACGAAGGT|GAGAAGAG/1 110 3 113 4
        # D00597:180:C7NMDANXX:6:1101:1184:35164-ACGAAGGT|GAGAAGAG/2 117 0 117 0

        tname1 = None
        tname2 = None
        while (tname1 != name1 or tname2 != name2):

            p1 = trim_log.readline().split()
            p2 = trim_log.readline().split()

            if len(p1) == 0 or len(p2) == 0:
                eof = True
                break

            if has_index:
                tname1, _, _, trimmed1_5p, last_base_pos1, trimmed1_3p = p1
                tname2, _, _, trimmed2_5p, last_base_pos2, trimmed2_3p = p2
            else:
                tname1, _, trimmed1_5p, last_base_pos1, trimmed1_3p = p1
                tname2, _, trimmed2_5p, last_base_pos2, trimmed2_3p = p2

            # prefix '@' to match name[12]
            tname1 = '@' + tname1
            tname2 = '@' + tname2

        if eof:
            # EOF
            break

        assert (tname1 == name1)
        assert (tname2 == name2)

        # e.g. "1^5"
        trimmed = "{}{}{}".format(trimmed1_5p, DELIM_ANNO_READ_PAIR,
                trimmed2_5p)

        if name1[-2:] == '/1':
            # Some datasets have a '/1' and '/2' at end of R1 and R2 read names
            # respectively.
            name = name1[:-2]
            # e.g. "ReadName-ACGT,AATT|GGCC;1^5"
            name_anno = DELIM_ANNO_TYPE.join((name, trimmed))
            record1 = "{}/1\n{}\n+\n{}\n".format(name_anno, seq1, qual1)
            record2 = "{}/2\n{}\n+\n{}\n".format(name_anno, seq2, qual2)
        elif has_index:
            # Some reads have the index included.
            name = name1
            name_anno = DELIM_ANNO_TYPE.join((name, trimmed))
            record1 = "{} {}\n{}\n+\n{}\n".format(name_anno, index_str1, seq1, qual1)
            record2 = "{} {}\n{}\n+\n{}\n".format(name_anno, index_str2, seq2, qual2)
        else:
            # Some reads have neither the index nor the /1, /2
            name = name1
            name_anno = DELIM_ANNO_TYPE.join((name, trimmed))
            record1 = "{}\n{}\n+\n{}\n".format(name_anno, seq1, qual1)
            record2 = "{}\n{}\n+\n{}\n".format(name_anno, seq2, qual2)

        # Write the files
        out1.write(record1)
        out2.write(record2)

def create_annotated_file(in1, trim_log, out1, has_index):
    """Creates single-end FASTQ file with read names annotated with amount of 5'-trimming.

    Example:
        "ReadName-ACGT" -> "ReadName-ACGT;1"

    Args:
        in1 (file): FASTQ input filehandle.
        trim_log (file): Trimmomatic trim log.
        out1 (file): Annotated fastq output filehandle.
        has_index (bool): Whether or not this FASTQ file has the index
            string (e.g. 'read_id 1:N:0:ATCACGTT')
    """
    eof = False
    while True:
        name1, seq1, _, qual1 = \
                (in1.readline().rstrip() for i in range(4))

        if not name1:
            break

        if has_index:
            name1, index_str1 = name1.split()

        # examples:
        # D00597:180:C7NMDANXX:6:1101:1184:35164-ACGAAGGT 110 3 113 4

        tname1 = None
        while (tname1 != name1):
            p1 = trim_log.readline().split()

            if len(p1) == 0:
                eof = True
                break

            if has_index:
                tname1, _, _, trimmed1_5p, last_base_pos1, trimmed1_3p = p1
            else:
                tname1, _, trimmed1_5p, last_base_pos1, trimmed1_3p = p1

            # prefix '@' to match name1
            tname1 = '@' + tname1

        if eof:
            # EOF
            break

        assert (tname1 == name1)

        # e.g. "1"
        trimmed = "{}".format(trimmed1_5p)

        if has_index:
            # Some reads have the index included.
            name = name1
            name_anno = DELIM_ANNO_TYPE.join((name, trimmed))
            record1 = "{} {}\n{}\n+\n{}\n".format(name_anno, index_str1, seq1, qual1)
        else:
            # Some reads have neither the index nor the /1, /2
            name = name1
            name_anno = DELIM_ANNO_TYPE.join((name, trimmed))
            record1 = "{}\n{}\n+\n{}\n".format(name_anno, seq1, qual1)

        # Write the files
        out1.write(record1)

def parse_args(args):
    """Parse the command line arguments."""

    # Convert ~ to real path (strip silly leading './' too)
    if args['<in2.fastq>']:
        # PE mode
        in1 = os.path.expanduser(args['<in1.fastq>'])
        in2 = os.path.expanduser(args['<in2.fastq>'])
        input_files = (in1, in2)
    else:
        fin = os.path.expanduser(args['<input.fastq>'])
        input_files = (fin, )

    # Figure out which function to use to write to output file.
    compress = args['--compress']
    num_threads = int(args['--threads'])
    if num_threads > 1 and which('pigz') and compress:
        # return a partial for pigzwrite
        write_func = functools.partial(pigzwrite, num_threads)
    else:
        write_func = functools.partial(open, mode = 'w')

    outdir = args_to_out_dir(args)

    # optional trimlog
    opt_trimlog = args['-l'] if args['-l'] else None

    return (write_func, outdir, compress, opt_trimlog, input_files)

def run(write_func, outdir, compress, opt_trimlog, input_files):
    """Start the run.

    Args:
        outdir (str): Output directory for results
        compress (bool): Whether or not to compress the output.
        input_files ([str]): Input files.
    """

    # Some fastq files have an index read in them. Detect.
    if is_gzipped(input_files[0]):
        with gzip.open(input_files[0], 'rb') as in1:
            first_line = in1.readline()
    else:
        with open(input_files[0], 'r') as in1:
            first_line = in1.readline()
    has_index = True if len(first_line.split()) > 1 else False

    if len(input_files) == 2:
        in1, in2 = input_files
        out1 = filename_in_to_out_fqgz(in1, SUFFIX_ANNOTATE_QTRIM, compress,
                outdir)
        out2 = filename_in_to_out_fqgz(in2, SUFFIX_ANNOTATE_QTRIM, compress,
                outdir)
        out_files = [out1, out2]

        if opt_trimlog is not None:
            trim_log = opt_trimlog
        else:
            qtrim_out2_file = out2.replace(SUFFIX_QTRIM + '.', '')
            trim_log = pe_log_filename(SUFFIX_QTRIM, in2, 'trimlog')

        tmp_out1, tmp_out2 = tmpf_start(out1, out2)

        with    pgopen(1, in1) as fin1, \
                pgopen(1, in2) as fin2, \
                pgopen(1, trim_log) as ftrim_log, \
                write_func(tmp_out1) as fout1, \
                write_func(tmp_out2) as fout2:
            create_annotated_files(fin1, fin2, ftrim_log, fout1, fout2, has_index)

        tmpf_finish(tmp_out1, tmp_out2)

    elif len(input_files) == 1:
        in1 = input_files[0]
        out1 = filename_in_to_out_fqgz(in1, SUFFIX_ANNOTATE_QTRIM, compress,
                outdir)
        out_files = [out1]

        if opt_trimlog is not None:
            trim_log = opt_trimlog
        else:
            qtrim_out_file = out1.replace(SUFFIX_QTRIM + '.', '')
            trim_log = se_log_filename(SUFFIX_QTRIM, in1, 'trimlog')

        tmp_out1 = tmpf_start(out1)[0]
        with    pgopen(1, in1) as fin1, \
                pgopen(1, trim_log) as ftrim_log, \
                write_func(tmp_out1) as fout1:
            create_annotated_file(fin1, ftrim_log, fout1, has_index)
        tmpf_finish(tmp_out1)

    else:
        raise ControlFlowException("""ERR911: Not possible to be here.""")

    return out_files


###############
### Classes ###
###############


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object
# py-lmdb uses bytes() for py3 and str() for py2.
# This package has different code for py2 and py3.
# And importing that future 'object' has a bug that screws up __slots__ in
# py2 (causes different behavior than in py3).

from builtins import range

# Dupligänger imports
try:
    from dupliganger.constants import *
    from dupliganger.exceptions import *
except ImportError:
    from constants import *
    from exceptions import *

# Other imports

import sys

# For progress and timing
import time

# For logging
import logging

# For pgopen, gzwrite, pigzopen
import contextlib
import os
import subprocess
import shlex

# For is_gzipped
import io

# For tmpfile names
import string
import random

# For bam
import gzip

# # For shell-like "which()"
try:
    from shutil import which
except ImportError:
    from whichcraft import which

# For memory usage.
import psutil
import resource

#################
### Functions ###
#################

def some_function(args):
    """A utility function."""

def pmsg(msg):
    """Dupligänger Message.  Write a message to STDOUT and the
    dupliganger.log.

    You should *never* use print(), sys.stderr.write() or sys.stdout.write()
    within Dupligänger (except for the class Progress).  Instead, use this function.

    """
    logging.info(msg)
    print(msg)

def perr(msg):
    """Dupligänger Error Message.  Same as pmsg(), but it's an error instead
    of info."""

    logging.error(msg)
    print(msg)

def fmt_time(seconds):
    """Format number of seconds into a human readable time string.

    Example:

        Example::

            3601.5 ->  "3601.5s (or 1h0m)"
            1000.5 ->  "1000.5s (or 16m40s)"
            59.5 ->  "59.5s"

    Arguments:
        seconds (float): A length of time given in seconds.

    """
    m, s = divmod(seconds, 60)
    h, m = divmod(int(m), 60)
    s = int(s)
    if h >= 1:
        return "{0:.1f}s (or {1}h{2}m)".format(int(seconds), h, m)
    elif m >= 1:
        return "{0:.1f}s (or {1}m{2}s)".format(seconds, m, s)
    else:
        return "{0:.1f}s".format(seconds)

def setup_logging(conf, time_start):
    """Setup dupliganger.log logging.

    Does the following:
        - configures the logging
        - adds a "header" line that includes the version and date
        - adds a line showing how Dupligänger was executed from the command line
        - dumps contents of dupliganger.config and samples_filelist to log

    Arguments:
        conf (Configuration): The Configuration singleton object.
        time_start (float): The time in seconds since the epoch.

    """
    # Configure logging
    logging.config.dictConfig(LOGGING)

    # First line of log, version + timedate.
    left = "Dupligänger! version {}.".format(__version__)
    right_len = 79 - len(left)
    pmsg("{}{:>{}}.".format(
        left, time.strftime('%l:%M:%S %p %Z on %b %d, %Y'), right_len))
    pmsg("")

    # Next, add how Dupligänger was run from the command line (only to the
    # log).
    args = sys.argv
    args[0] = os.path.basename(sys.argv[0])
    logging.info("Excecuted as:\n")
    logging.info("    {}".format(" ".join(args)))

    # Next, add the full contents of the configuration file.
    header = "=== Configuration File '{}' Contents ===".format(
        conf.general.config_file)
    logging.info("")
    logging.info("{:^79}".format(header))
    logging.info("")
    with open(os.path.expanduser(conf.general.config_file), 'r') as f:
        logging.info(f.read())

    # Next, add the full contents of the samples_filelist.
    header = "=== Samples Filelist '{}' Contents ===".format(
        conf.general.samples_filelist)
    logging.info("")
    logging.info("{:^79}".format(header))
    logging.info("")
    with open(os.path.expanduser(conf.general.samples_filelist), 'r') as f:
        logging.info(f.read())

    # Finally, add a header line for Dupligänger exection.
    header = "=== Dupligänger Execution ==="
    logging.info("{:^79}".format(header))
    logging.info("")


@contextlib.contextmanager
def pgopen(num_threads, filename):
    """Context manager to open a plain or gzipped file for reading only.

    Read-only.  If num_threads > 1 and pigz is installed, it will use pigz to
    uncompress.

    Usage:
        with pgopen(file) as f:
            # do stuff...
    """
    # Determine if gzipped
    gzipped = False
    with io.open(filename, mode='rb') as f:
        # Is this a gzipped file? Check magic number.
        magic_num = f.read(2)
        f.seek(0)

        if magic_num == b'\x1f\x8b':
            gzipped = True

    if gzipped:
        # This is a gzipped file.
        if num_threads > 1 and which('pigz'):
            # pigz read
            cmd = shlex.split('unpigz -p {} -c {}'.format(num_threads, filename))
            with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=-1,
                    universal_newlines=True).stdout as f:
                yield f
        else:
            # gzip read
            cmd = shlex.split('gunzip -c {}'.format(filename))
            with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=-1,
                    universal_newlines=True).stdout as f:
                yield f
    else:
        # This is not a gzipped file.
        with asciiopen(filename) as f:
            yield f

@contextlib.contextmanager
def asciiopen(filename):
    """Context manager to open an ASCII encoded file across Python 2/3.

    Usage:
        with asciiopen(file) as f:
            # do stuff...
    """
    # So, basically, you came to the conclusion that you probably just
    # want to write a function 'asciiopen' which just opens up a file
    # differently depending on whether you're running python 2 or 3.
    # Why?
    # I fought with io.open a bunch, and here's how it behaves:
    #
    # Python 2:
    #   >>> type(io.open('ACTTGATG_R1.3300.fq', mode='r', encoding='ascii').readline())
    #   <type 'unicode'>
    # Python 3
    #   >>> type(io.open('ACTTGATG_R1.3300.fq', mode='r', encoding='ascii').readline())
    #   <class 'str'>
    #
    # The alternative is to read everything in as bytes, but python 3 is
    # weird about that as well:
    #   >>> a = io.open('ACTTGATG_R1.3300.fq', mode='rb')
    #   >>> b = a.readline()
    #   >>> b
    #   b'@D00597:180:C7NMDANXX:6:1101:1184:33231/1\n'
    #   >>> ascii(b)
    #   "b'@D00597:180:C7NMDANXX:6:1101:1184:33231/1\\n'"
    #
    # In the end, it's probably just easier to write an asciiopen()
    # function where in python3 it is:
    #   io.open('ACTTGATG_R1.3300.fq', mode='r', encoding='ascii'
    # and in python2 it is just the old "open".  Definitely do NOT do:
    #       from builtins import open
    #
    # There is perhaps a better way to do all this.  For now, this works.
    #
    if sys.version_info.major == 2:
        with open(filename, 'r') as f:
            yield f
    elif sys.version_info.major == 3:
        with io.open(filename, mode='r', encoding='ascii') as f:
            yield f

@contextlib.contextmanager
def bamopen(filename, silence_broken_pipe_errors=False):
    """Context manager to open a BAM file for reading only.

    Usage:
        with bamopen(file) as f:
            # do stuff...

    Args:
        filename (str): The filename to examine.
        silence_broken_pipe_errors (bool): If True, then pipe stderr to
            /dev/null.  This is useful if you only want to "peek" at the first
            few bytes / lines of a file; in that scenario, if you close before
            reading all the output, samtools will complain of a broken pipe.
    """

    cmd = shlex.split('samtools view -h {}'.format(filename))
    if silence_broken_pipe_errors:
        import os
        DEVNULL = open(os.devnull, 'wb')
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=-1,
                stderr=DEVNULL, universal_newlines=True).stdout as f:
            yield f
    else:
        with subprocess.Popen(cmd, stdout=subprocess.PIPE,
                bufsize=-1, universal_newlines=True).stdout as f:
            yield f

@contextlib.contextmanager
def sambamopen(filename):
    """Context manager to open a SAM or BAM file for reading only.

    Usage:
        with sambamopen(file) as f:
            # do stuff...

    Args:
        filename (str): The filename to examine.
    """

    cmd = shlex.split('samtools view -h {}'.format(filename))
    if is_bam(filename):
        with subprocess.Popen(cmd, stdout=subprocess.PIPE,
                bufsize=-1, universal_newlines=True).stdout as f:
            yield f
    else:
        with pgopen(1, filename) as f:
            yield f

@contextlib.contextmanager
def gzwrite(filename):
    """Context manager to write to a gzipped file.

    Usage:
        with gzwrite(file) as f:
            # do stuff...
    """
    with open(filename, 'wb') as f:
        cmd = shlex.split('gzip -c')
        z = subprocess.Popen(cmd, stdin=subprocess.PIPE, bufsize=-1, stdout=f)
        try:
            yield z.stdin
        finally:
            z.stdin.close()
            z.wait()

@contextlib.contextmanager
def pigzwrite(num_threads, filename):
    """Context manager to write to a gzipped file with pigz.

    Usage:
        with gzwrite(num_threads, file) as f:
            # do stuff...
    """
    with open(filename, 'wb') as f:
        cmd = shlex.split('pigz -p {}'.format(num_threads))
        z = subprocess.Popen(cmd, stdin=subprocess.PIPE, bufsize=-1, stdout=f)
        try:
            yield z.stdin
        finally:
            z.stdin.close()
            z.wait()

def file_root(filename, expected_ext=None):
    """Given a filename, returns the root name to be used for new files.

    If the file is gzipped, strips that away first.

    Arguments:
        filename (str): The filename to work on.
        expected_ext ([str]): An array of expected filename extensions.  Will
            raise exception if doesn't match.

    >>> file_root('filename.fq')
    ('', 'filename', '.fq')
    >>> file_root('filename.fq.gz')
    ('', 'filename', '.fq.gz')
    >>> file_root('/path/to/filename.fq')
    ('/path/to', 'filename', '.fq')
    >>> file_root('filename.fq.gz', ['fq', 'fastq'])
    ('', 'filename', '.fq.gz')
    >>> file_root('filename.log', ['fq', 'fastq'])
    Traceback (most recent call last):
        ...
    CannotContinueException: Extension log was in not expected extensions ['fq', 'fastq']
    """

    dirname, basename = os.path.split(os.path.expanduser(filename))
    root, ext = os.path.splitext(basename)
    orig_ext = ext
    if ext in ('.gz', ):
        root, ext = os.path.splitext(root)
        orig_ext = ext + orig_ext
    ext = ext[1:]

    if expected_ext:
        if ext not in expected_ext:
            raise UnexpectedExtensionException(
                    "Extension {} was in not expected extensions {}".format(
                        ext, expected_ext))

    return (dirname, root, orig_ext)

def pe_file_root(pe_filename):
    """Given a paired-end filename, returns the root name to be used for new files.

    Will raise an exception if passed an R1 filename (why? so we don't
    accidentally use this function for single-end reads).

    If the file is gzipped, strips that away first.

    Arguments:
        pe_filename (str): The R2 paired end filename.

    Returns:
        (str, str, str): A 3-tuple: (the path, the pe_file_root, the original
                extension)

    >>> pe_file_root('/path/to/filename_R2.fq')
    ('/path/to', 'filename', '.fq')
    >>> pe_file_root('/path/to/filename_R2.fq.gz')
    ('/path/to', 'filename', '.fq.gz')
    >>> pe_file_root('/path/to/filename_R2_other_stuff.fq.gz')
    ('/path/to', 'filename_other_stuff', '.fq.gz')

    """
    dirname, root, orig_ext = file_root(pe_filename, ['fq', 'fastq'])

    if 'R1' in root or 'R2' not in root:
        raise CannotContinueException(
                 "pe_file_root should only be called with the R2 file.")

    if 'R2' in root.replace('R2', ''):
        m = "'R2' occurs more than once in {}, please change filename".format(
                 pe_filename)
        raise CannotContinueException(m)

    root = root.replace('_R2', '')
    return (dirname, root, orig_ext)

def se_log_filename(prefix, se_filename, suffix='out'):
    """Given a single-end (R1) filename, returns an appropriate log filename to
    be used for logging output of external tools.

    >>> se_log_filename('rmadapt', 'filename.fq')
    'exec.rmadapt.sr.filename_out'
    >>> se_log_filename('rmadapt', 'filename.fq.gz')
    'exec.rmadapt.sr.filename_out'
    >>> se_log_filename('rmadapt', 'filename_other_stuff.fq.gz')
    'exec.rmadapt.sr.filename_other_stuff.out'
    >>> se_log_filename('rmadapt', '/path/to/filename.fq')
    '/path/to/exec.rmadapt.sr.filename.out'

    """
    dirname, root, junk = file_root(se_filename, ['fq', 'fastq'])
    if not dirname or dirname == '.':
        return "exec.{}.sr.{}.{}".format(prefix, root, suffix)
    else:
        return "{}/exec.{}.sr.{}.{}".format(dirname, prefix, root, suffix)

def pe_log_filename(prefix, pe_filename, suffix='out'):
    """Given a paired-end (R2) filename, returns an appropriate log filename to
    be used for logging output of external tools.

    >>> pe_log_filename('rmadapt', 'filename_R2.fq')
    'exec.rmadapt.pe.filename.out'
    >>> pe_log_filename('rmadapt', 'filename_R2.fq.gz')
    'exec.rmadapt.pe.filename.out'
    >>> pe_log_filename('rmadapt', 'filename_R2_other_stuff.fq.gz')
    'exec.rmadapt.pe.filename_other_stuff.out'
    >>> pe_log_filename('rmadapt', '/path/to/filename_R2.fq')
    '/path/to/exec.rmadapt.pe.filename.out'

    """
    dirname, root, ext = pe_file_root(pe_filename)
    if not dirname or dirname == '.':
        return "exec.{}.pe.{}.{}".format(prefix, root, suffix)
    else:
        return "{}/exec.{}.pe.{}.{}".format(dirname, prefix, root, suffix)

def filename_in_to_out_fqgz(filename, suffix, gzip, outdir):
    """Converts a FASTQ filename to new filename with suffix, optionally gzipped.

    Examples:
        filename_in_to_out_fqgz('/path/to/file.fq.gz', 'stage1')
            -> ./file.stage1.fq.gz
        filename_in_to_out_fqgz('/path/to/file.fq.gz', 'stage1', False)
            -> ./file.stage1.fq
        filename_in_to_out_fqgz('/path/to/file.fq', 'stage1')
            -> ./file.stage1.fq.gz
        filename_in_to_out_fqgz('/path/to/file.fq.gz', 'stage1', True, 'outdir')
            -> outdir/file.stage1.fq.gz

    Args:
        filename (str): Name of file to transform.
        suffix (str): Suffix to add to filename.
        gzip (bool): Whether or not to add a "gz" on the end.
        outdir (Optional[str]): Directy to place filename in.

    Returns:
        str: The new filename.
    """
    dirname, root, junk = file_root(filename, ('fq', 'fastq'))
    ext = 'fq.gz' if gzip else 'fq'
    f = "{}.{}.{}".format(root, suffix, ext)
    return os.path.join(outdir, f)

def filename_in_to_out_sambam(filename, suffix, outdir):
    """Converts a BAM or SAM filename to new output filename with suffix.

    Args:
        filename (str): Name of file to transform.
        suffix (str): Suffix to add to filename.
        outdir (str): Directy to place filename in.

    Returns:
        str: The new filename.

    >>> filename_in_to_out_sambam('/path/to/file', 'dups_removed.sam', 'outdir')
    'outdir/file.dups_removed.sam'
    >>> filename_in_to_out_sambam('/path/to/file.sam', 'dups_removed.sam', 'outdir')
    'outdir/file.dups_removed.sam'
    >>> filename_in_to_out_sambam('/path/to/file.bam', 'dups_removed.sam', 'outdir')
    'outdir/file.dups_removed.sam'
    >>> filename_in_to_out_sambam('/path/to/file', 'dups_flagged.sam', 'outdir')
    'outdir/file.dups_flagged.sam'
    """
    try:
        dirname, root, junk = file_root(filename, ('sam', 'bam'))
    except(UnexpectedExtensionException):
        # Doesn't look like this file has a bam or sam extension
        dirname, root = os.path.split(os.path.expanduser(filename))

    f = "{}.{}".format(root, suffix)
    return os.path.join(outdir, f)

def filename_in_bam_to_out_fqgz(filename, suffix, gzip, paired, outdir):
    """Converts a BAM filename to new FASTQ single-end OR paired-end
    filename(s) with suffix, optionally gzipped.

    Args:
        filename (str): Name of file to transform.
        suffix (str): Suffix to add to filename.
        gzip (bool): Whether or not to add a "gz" on the end.
        paired (bool): True: paired-end data, False, single-end data.
        outdir (str): Directory to place filename in.

    Returns:
        [str]: The new filename(s).

    >>> filename_in_bam_to_out_fqgz('/path/to/file.bam', 'stage1', True, True, '.')
    ['./file_R1.stage1.fq.gz', './file_R2.stage1.fq.gz']
    >>> filename_in_bam_to_out_fqgz('/path/to/file.bam', 'stage1', False, True, '.')
    ['./file_R1.stage1.fq', './file_R2.stage1.fq']
    >>> filename_in_bam_to_out_fqgz('/path/to/file.bam', 'stage1', True, True, 'outdir')
    ['outdir/file_R1.stage1.fq.gz', 'outdir/file_R2.stage1.fq.gz']
    >>> filename_in_bam_to_out_fqgz('/path/to/file.bam', 'stage1', True, False, '.')
    ['./file.stage1.fq.gz']
    >>> filename_in_bam_to_out_fqgz('/path/to/file.bam', 'stage1', False, False, '.')
    ['./file.stage1.fq']
    >>> filename_in_bam_to_out_fqgz('/path/to/file.bam', 'stage1', True, False, 'outdir')
    ['outdir/file.stage1.fq.gz']
    """
    dirname, root, junk = file_root(filename, ('bam', 'BAM'))
    ext = 'fq.gz' if gzip else 'fq'
    if paired:
        f1 = "{}_R1.{}.{}".format(root, suffix, ext)
        f2 = "{}_R2.{}.{}".format(root, suffix, ext)
        fqgzs = [os.path.join(outdir, f1), os.path.join(outdir, f2)]
    else:
        f = "{}.{}.{}".format(root, suffix, ext)
        fqgzs = [os.path.join(outdir, f)]

    return fqgzs

def args_to_out_dir(args):
    """Convenience function to retrieve the output directory from args.

    Also mkdir the out_dir if it doesn't exist.

    Arguments:
        args (dict): Arguments from docopt.

    Returns:
        str: The name of the output directory.
    """

    if '-o' in args and args['-o'] is not None:
        out_dir = args['-o']
    elif '--out' in args and args['--out'] is not None:
        our_dir = args['--out']
    elif '--outdir' in args and args['--outdir'] is not None:
        our_dir = args['--outdir']
    elif '--out-dir' in args and args['--out-dir'] is not None:
        our_dir = args['--out-dir']
    else:
        out_dir = ""

    # Convert ~ to real path
    out_dir = os.path.expanduser(out_dir)

    try:
        os.mkdir(out_dir)
    except OSError:
        try:
            os.makedirs(out_dir)
        except OSError:
            pass

    # Confirm existence if not ""
    if out_dir != '' and not os.path.exists(out_dir):
        m = "Unable to create output directory {}".format(out_dir)
        raise CannotContinueException(m)

    return out_dir

def tmpf_start(*filenames):
    """Takes a list of filenames and returns "temporary" versions of those
    filenames. Works in conjuction with tmpf_finish().

    Handles gzipped files as well.

    Note: No actual files are created; only the names of files are created.

    Args:
        filename ([str]): The filename(s).
    Returns:
        [str]: A temporary version of the filename(s).
    """
    tmp_filenames = []
    for filename in filenames:
        if filename == os.devnull:
            # If you want /dev/null, it's not very temporary in nature.
            tmp_filenames.append(os.devnull)
            continue
        is_gzipped = False
        if filename[-3:] == '.gz':
            is_gzipped = True
            filename = filename[:-3]
        chars = string.ascii_lowercase + string.digits
        size = TMP_FILE_NAME_RANDOM_STR_SIZE
        random_str = ''.join(random.choice(chars) for _ in range(size))
        tmp_filename = "{}.tmp.{}".format(filename, random_str)
        if is_gzipped:
            tmp_filename += '.gz'
        tmp_filenames.append(tmp_filename)
    return tmp_filenames

def tmpf_finish(*tmp_filenames):
    """Makes a temporary version of a file permanent.  It strips off the
    '.tmp.XXXXXX' part of the filename, then renames (moves) the actual file.
    Does no checking to see if the file is still open (hopefully it isn't).

    Handles gzipped files as well.

    Args:
        tmp_filenames ([str]): The temporary name of the file(s).
    """
    final_filenames = []
    for tmp_filename in tmp_filenames:
        if tmp_filename == os.devnull:
            # If you want /dev/null, it's not very temporary in nature.
            continue
        is_gzipped = False
        if tmp_filename[-3:] == '.gz':
            is_gzipped = True
            tmp_filename = tmp_filename[:-3]
        size = len(".tmp.") + TMP_FILE_NAME_RANDOM_STR_SIZE
        filename = tmp_filename[:-size]
        ext = tmp_filename[-size:]
        assert(ext[0:5] == '.tmp.')
        if is_gzipped:
            tmp_filename += '.gz'
            filename += '.gz'
        final_filenames.append(filename)
        os.rename(tmp_filename, filename)
    return final_filenames

@contextlib.contextmanager
def tmpf_name(delete_temp_files_upon_failure, filename):
    """Context manager to manage naming of temporary files (for writing).  Does the
    following:
        * creates *temporary* filename
        * yields that filename back
        * Upon success, renames the temp file to permanent name ('filename').
        * Upon failure (an Exception), it will attempt to delete the temporary
          file.
    Args:
        delete_temp_files_upon_failure (bool): Whether or not to delete tmp
            files upon external failure.
        filename (str): Name of file to be written to.
    Yields:
        (str): The name of the temp file to be written to.
    """
    try:
        tmp_filename = tmpf_start(filename)[0]
        yield tmp_filename
    except Exception as e:
        if delete_temp_files_upon_failure:
            try:
                os.remove(tmp_filename)
            except Exception:
                # We catch and ignore this exception so as to not mask previous exceptions.
                pass
        raise e
    finally:
        try:
            tmpf_finish(tmp_filename)
        except Exception:
            # We catch and ignore this exception so as to not mask previous
            # exceptions.
            pass

@contextlib.contextmanager
def tmpf_open(fp_write, delete_temp_files_upon_failure, filename):
    """Context manager to open up temporary files for writing.  Does the
    following:
        * opens *temporary* output file for writing
        * yields that file back for writing
        * when finished writing, closes the file.
        * Upon success, renames the temp file to permanent name ('filename').
        * Upon failure (an Exception), it will attempt to delete the temporary
          file.
    Args:
        fp_write (function): Function to write file (eg write(), gzwrite()).
        delete_temp_files_upon_failure (bool): Whether or not to delete tmp
            files upon external failure.
        filename (str): Name of file to be written to.
    Yields:
        (file): A filehandle to the temp file.
    """
    try:
        tmp_filename = tmpf_start(filename)[0]
        with fp_write(tmp_filename) as f:
            yield f
    except Exception as e:
        if delete_temp_files_upon_failure:
            try:
                os.remove(tmp_filename)
            except Exception:
                # We catch and ignore this exception so as to not mask previous exceptions.
                pass
        raise e
    finally:
        try:
            tmpf_finish(tmp_filename)
        except Exception:
            # We catch and ignore this exception so as to not mask previous exceptions.
            pass

def is_gzipped(filename):
    """Determines if the file is gzipped or not. Uses magic number.

    Args:
        filename (str): The file to examine.
    Returns:
        bool: True if it is a gzipped file, False otherwise.
    """
    with io.open(filename, mode='rb') as f:
        # Is this a gzipped file? Check magic number.
        magic_num = f.read(2)

    return magic_num == b'\x1f\x8b'

def is_bam(filename):
    """Determines if the file is a BAM file or not. Uses magic number.

    Args:
        filename (str): The file to examine.
    Returns:
        bool: True if it is a BAM file, False otherwise.
    """
    try:
        with gzip.open(filename, 'rb') as f:
            magic_num = f.read(3)
            if magic_num == b'BAM':
                return True
            else:
                return False
    except IOError:
        # If it's not gzipped, then it's not a BAM.
        return False

def is_paired_bam(filename):
    """Detects if a BAM file is paired.

    Args:
        filename (str): The file to examine.
    Returns:
        bool: True if it is a paired BAM file, False otherwise.
    """
    assert(is_bam(filename))

    name1 = None
    name2 = None

    with bamopen(filename, True) as f:
        name1 = f.readline().split()[0]
        name2 = f.readline().split()[0]

    return name1 == name2

def setup_report_db():
    """Sets up a key-value pair "Report Database" to track metrics."""

    report_db = {}
    for metric in REPORT_DB_COUNT_METRICS:
        report_db[metric] = 0
    for metric in REPORT_DB_8_UMI_DIST_COUNT_METRICS:
        for i in range(1,9):
            report_db["{}{}".format(metric, i)] = 0

    return report_db

def memory_info(report_in_MBytes=False):
    """Report current memory used and maximum RSS memory consumed.  Results are
    in 'bytes' by default; passing 'report_in_MBytes' returns results in
    megabytes.

    Returns:
        (float, float): A two-tuple.  First member is current memory used,
            second member is maximum memory used.
    """

    # Currently used memory.
    process = psutil.Process(os.getpid())
    try:
        # py2
        curr_mem = process.get_memory_info().rss
    except(AttributeError):
        # py3
        curr_mem = process.memory_info().rss

    # Maximum memory (RSS) used.
    if sys.platform == 'darwin':
        max_mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    else:
        max_mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024

    if report_in_MBytes:
        return (curr_mem / 1024**2, max_mem / 1024**2)
    else:
        return (curr_mem, max_mem)

###############
### Classes ###
###############

class SomeSharedClass(object):
    """Some shared class that's useful."""
    pass

class Progress(object):
    """A simple progress meter printing to STDERR and dupliganger.log.

    We print to STDERR just for convenience because it's not buffered.

    """
    def __init__(self, progress_str, step=None, known_total=None, **kwargs):
        self.start = time.time()
        self.count = 0
        self.total = 0
        self.progress_str = progress_str
        self.known_total = None if known_total is None else int(known_total)
        self.step = None if step is None else int(step)
        self.indent = kwargs.get('indent', 0) * ' '
        sys.stderr.write("{}{}...".format(self.indent, progress_str))
        if self.step:
            sys.stderr.write('\n')

    def progress(self):
        if self.step:
            self.count += 1
            self.total += 1
            if self.count == self.step:
                if self.known_total:
                    sys.stderr.write("\r{}    {}/{} so far...".format(
                        self.indent, self.total, self.known_total))
                else:
                    sys.stderr.write("\r{}    {} so far...".format(
                        self.indent, self.total))
                self.count = 0
        else:
            sys.stderr.write('.')

    def done(self):
        _end = time.time()
        _secs = _end - self.start

        if not self.step:
            msg_mid = "{}{}...".format(self.indent, self.progress_str)
            sys.stderr.write('\r')
            sys.stderr.write(msg_mid)
            msg_mid = ""
        else:
            if self.known_total:
                msg_mid = "{}    {}/{} so far...".format(
                    self.indent, self.total, self.known_total)
                sys.stderr.write('\r')
                sys.stderr.write(msg_mid)
            else:
                msg_mid = "{}    {} so far...".format(self.indent, self.total)
                sys.stderr.write('\r')
                sys.stderr.write(msg_mid)

        msg_end = " done. (elapsed time: {})".format(fmt_time(_secs))
        sys.stderr.write(msg_end + '\n')

        # Now write full message to the log file.

        msg_beg = ("{}{}...".format(self.indent, self.progress_str))
        if self.step:
            msg_beg += '\n'
        logging.info(msg_beg + msg_mid + msg_end)

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Dupligänger

A reference-based, UMI-aware, 5'-trimming-aware PCR duplicate removal pipeline.

Usage: dupliganger [options] <command> [<args>...]


Dupligänger is a pipeline.  Each stage of the pipeline is run by passing a
'command' to Dupligänger.  The commands / pipeline-steps (in order) are as
follows:

   remove-umi       1. Annotate read names with UMIs (clip inline UMIs if needed).
   remove-adapter   2. Remove adapters ('Cutadapt' wrapper).
   qtrim            3. Quality trim ('Trimmomatic' wrapper).
   annotate-qtrim   4. Annotates quality trimmed file(s).
   align            5. Align reads to a reference genome assembly. -- NOT YET IMPLEMENTED
   dedup            6. Use the alignment to remove PCR duplicates.

While generally used only by the developers of Dupligänger, the 'dedup'
command is comprised of the following Dupligänger commands run in the
following order:

    build-read-db       1. Build a database of aligned reads.
    build-location-db   2. Build a database of locations of aligned reads.
    build-dup-db        3. Build a database of PCR duplicates.

Options:
    -o OUT_DIR      Place results in directory OUT_DIR.
    --compress      Compress output.

Note:
    Dupligänger supports (and autodetects) input FASTQ files that are gzipped.

See 'dupliganger help <command>' for more information on a specific command.

"""

# Python 3 imports
from __future__ import absolute_import
from __future__ import division

# For Docopt
from docopt import docopt
import importlib

# version
try:
    from dupliganger._version import __version__
except ImportError:
    from _version import __version__
__version_info__ = tuple(__version__.split('.'))

# Other imports
import subprocess


#################
### Constants ###
#################

SDD_COMMAND_MODULES = ['remove_umi', 'barcode_split_quality_filter_umi_anno',
        'remove_adapter', 'qtrim', 'annotate_qtrim', 'prep', 'dedup',
        'build_read_and_loc_dbs', 'build_dup_db']

#################
### Functions ###
#################


##############################################
### Config File and Command Line Arguments ###
##############################################


###############
### Classes ###
###############


#######################
### Debug functions ###
#######################


############
### Main ###
############

def main():

    args = docopt(__doc__,
          version='dupliganger version {}'.format(__version__),
          options_first=True)

    cmd = args['<command>'].replace('-', '_')

    if cmd in SDD_COMMAND_MODULES:
        cmd_module = importlib.import_module("dupliganger.{}".format(cmd))
        cmd_module.main()
    elif cmd in ['help', None]:
        argv = args['<args>']
        if argv == []:
            exit(subprocess.call(['dupliganger', '--help']))
        elif argv[0].replace('-', '_') in SDD_COMMAND_MODULES:
            exit(subprocess.call(['dupliganger', argv[0], '--help']))
        else:
            exit("{} is not a dupliganger command. See 'dupliganger help'.".format(
                argv[0]))
    else:
        exit("{} is not a dupliganger command. See 'dupliganger help'.".format(args['<command>']))


if __name__ == "__main__":
    main()

# vim: softtabstop=4:shiftwidth=4:expandtab



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

# This file allows you to run dupliganger (for development purposes) like so:
#   git clone ...dupliganger
#   cd dupliganger
#   python -m dupliganger

# Python 3 imports
from __future__ import absolute_import
from __future__ import division

# Dupligänger imports
from dupliganger import dupliganger

def main():
    dupliganger.main()

if __name__ == "__main__":
    main()

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object

from future import standard_library
standard_library.install_aliases()

# Dupligänger imports
from dupliganger.build_read_and_loc_dbs import __doc__
from dupliganger.build_read_and_loc_dbs import run, parse_args

# Dupligänger test library imports
from common import (fix_paths, test_outdir, to_eout_db_dump_filename,
        BASE_OUTFILE_DIR, BASE_FILES_DIR)

## Other imports
import pytest
from docopt import docopt
import os


#################
### Constants ###
#################

# e.g. 'test_build_read_and_loc_dbs' -> 'build_read_and_loc_dbs'
COMMAND_BEING_TESTED = __name__[5:]

# e.g. 'dupliganger/test/files/last_run/build_read_and_loc_dbs'
OUTFILE_DIR = os.path.join(BASE_OUTFILE_DIR, COMMAND_BEING_TESTED)

# e.g. 'dupliganger/test/files/build_read_and_loc_dbs/in'
INFILE_DIR = os.path.join(BASE_FILES_DIR, COMMAND_BEING_TESTED, 'in')

# Taken from the docstring of build-read-and-loc-dbs
COMMAND_FILE_PARAMS = ['<alignment-file>']


#################
### Functions ###
#################

@pytest.fixture(scope="module", params=[
        'build-read-and-loc-dbs --store memory 11_first_test.pe.sam',
        'build-read-and-loc-dbs --store lmdb 11_first_test.pe.sam',
        'build-read-and-loc-dbs --store memory 12_one_dup_one_not.pe.sam',
        'build-read-and-loc-dbs --store lmdb 12_one_dup_one_not.pe.sam',
        'build-read-and-loc-dbs --store memory 12_one_dup_one_not.pe.bam',
        'build-read-and-loc-dbs --store lmdb 12_one_dup_one_not.pe.bam',
    ])
def commands(request):
    """This is where the magic happens.

    This is a pytest fixture that is parameterized.  If you pass in this
    fixture to a given test, that test will be run once for each param in
    params.

    In this fixture, we are passing commandline strings to dupliganger
    build-read-and-loc-dbs.

    See also "fix_paths()", which converts the simple filenames above to
    relative paths appropriate for this.
    """
    yield request.param


class TestBuildReadAndLocDbs(object):
    """Test build-read-and-loc-dbs."""

    #############
    ### Tests ###
    #############

    def test_run_all_fixture_commands(self, test_outdir, commands):
        """This test runs all the commands parameterized in the 'commands'
        fixture defined above.  (pytest is actually responsible for running the
        running this test once for each command / fixture parameter).

        Args:
            test_outdir (a fixture): This is the output directory into which
                all output files are placed.
            commands (a parameterized fixture): This boils down to actual
                command line arguments given to dupliganger.
        """

        # Get the arguments passed
        args = docopt(__doc__, help=False, argv=commands.split())
        # Fix input file path(s) and outdir path (prepend INFILE_DIR and OUTFILE_DIR)
        args = fix_paths(args, INFILE_DIR, OUTFILE_DIR, COMMAND_FILE_PARAMS)
        # Get alignment file...
        alignment_file = args['<alignment-file>']
        # Run dupliganger command!
        (parent_db, read_group_db, loc_db) = run(*parse_args(args))

        # 13_third_test.sam -> 13_third_test.read_group_db
        # 13_third_test.sam -> 13_third_test.read_name_db
        # 13_third_test.sam -> 13_third_test.loc_db

        for db_name in 'read_group_db loc_db'.split():
            # Get string representation of db
            db_repr = str(locals()[db_name])
            # Get expected out filename
            eout = to_eout_db_dump_filename(alignment_file, db_name)
            with open(eout, 'r') as e:
                assert db_repr == e.read()

        # assert 0

# vim: softtabstop=4:shiftwidth=4:expandtab




# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

##################
### Exceptions ###
##################

class ConfigurationException(Exception):
    """Something went wrong with the configuration section."""
class ExecutionException(Exception):
    """Something went wrong with the execution of an external command."""
class ControlFlowException(Exception):
    """Something went wrong with control flow logic."""
class CannotContinueException(Exception):
    """Dupligänger has encountered a situation from which it cannot continue."""
class PrerequisitesException(Exception):
    """Dupligänger is missing prerequisites (e.g. missing gsnap)."""
class ArgumentException(Exception):
    """Problem with command line arguments."""
class ParseException(Exception):
    """Something went wrong with parsing something."""
class HardClippingNotSupportedException(Exception):
    """Encountered sam file hard-clipping, which is not supported."""
class UnexpectedExtensionException(Exception):
    """Extension passed was not in expected extensions."""

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Annotates read names with UMIs and clips inline UMIs if needed.

Usage:
    dupliganger remove-umi [options] <input.fastq>
    dupliganger remove-umi [options] <in1.fastq> <in2.fastq>
    dupliganger remove-umi [options] <input.bam>


Note:
    Dupligänger supports (and autodetects) input FASTQ files that are gzipped.

Note:
    If passing a paired-end BAM file, it needs to be sorted by read name
    (if not sorted, dupliganger will exit out when it detects mismatching
    adjacent records).

Options:
    -h, --help
    -v, --verbose         Be verbose
    -k KIT, --kit KIT     The kit used [default: bioo].
    -o OUT_DIR            Place results in directory OUT_DIR.
    --compress            Compress (gzip) output files.
    -t N, --threads N     EXPERIMENTAL: If pigz is installed and --threads is
                          specified, output FASTQ files will be compressed with
                          pigz -p <n>; otherwise, they will be left
                          uncompressed (as it simply takes too long to compress
                          with just gzip) [default: 1].
    --force-paired        Do not autodetect whether paired-end vs single-end,
                          instead force paired-end. Helpful fail safe if you
                          believe you have paired end data.
"""

###############
### Imports ###
###############

# Python 3 imports
from __future__ import absolute_import
from __future__ import division
from builtins import range

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *
from dupliganger.common import (pgopen, bamopen, gzwrite, pigzwrite,
        tmpf_start, tmpf_finish, is_gzipped, is_bam, is_paired_bam,
        filename_in_to_out_fqgz, filename_in_bam_to_out_fqgz, args_to_out_dir)

## Other imports
from docopt import docopt

# For filename fixing
import os

# For writing gzipped files
import gzip

# For checking presence of illumina index under python 2/3.
import io

# For shell-like "which()"
try:
    from shutil import which
except ImportError:
    from whichcraft import which

# For function partials
import functools


#################
### Constants ###
#################


#################
### Functions ###
#################

def create_annotated_file_from_bam(fp_extract_umi, in1, out1):
    """Using a single-end BAM file as input, convert to FASTQ output file
    with read name annotated with UMI.

    Args:
        fp_extract_umi (function): A function used to extract the UMI.
        in1 (file): Single-end BAM input file.
        out1 (file): Annotated output fastq file.
    """

    # Walk the file and create a new annotated file
    while True:
        line1 = in1.readline()
        if not line1:
            # EOF
            break

        parts1 = line1.split()
        name1, seq1, qual1 = parts1[0], parts1[9], parts1[10]

        name1 = '@' + name1

        clip_len1, umi1 = fp_extract_umi(seq1)
        umi_anno = umi1

        name1_anno = "{}{}{}".format(name1, DELIM_ANNO, umi_anno)

        out1.write("{}\n{}\n+\n{}\n".format(name1_anno, seq1[clip_len1:],
            qual1[clip_len1:]))

def create_annotated_files_from_bam(fp_extract_umis, in1, out1, out2):
    """Using a paired-end BAM file as input, create paired FASTQ output files
    with read name annotated with UMI.

    Args:
        fp_extract_umis (function): A function used to extract the UMIs.
        in1 (file): Paired-end BAM input file.
        out1 (file): Read1 annotated output fastq file.
        out2 (file): Read2 annotated output fastq file.
    """

    # Walk the file and create a new annotated file
    while True:
        line1 = in1.readline()
        line2 = in1.readline()
        if not line1:
            # EOF
            break

        parts1 = line1.split()
        parts2 = line2.split()
        name1, seq1, qual1 = parts1[0], parts1[9], parts1[10]
        name2, seq2, qual2 = parts2[0], parts2[9], parts2[10]

        name1 = '@' + name1
        name2 = '@' + name2

        clip_len1, clip_len2, umi1, umi2 = fp_extract_umis(seq1, seq2)
        umis_anno = DELIM_ANNO_READ_PAIR.join([umi1, umi2])

        # Confirm that the file has been sorted!!!
        if name1 != name2:
            raise CannotContinueException(
                     """Expecting paired-end BAM file, but it does not appear
                     to be sorted by name.""")

        # Some datasets have a '/1' and '/2' at end of R1 and R2 read names
        # respectively.
        if name1[-2:] == '/1':
            neg_index = -2
            name = name1[:-2]
            name_anno = "{}{}{}".format(name, DELIM_ANNO, umis_anno)
            name1_anno = "{}/1".format(name_anno)
            name2_anno = "{}/2".format(name_anno)
        else:
            neg_index = 0
            name1_anno = "{}{}{}".format(name1, DELIM_ANNO, umis_anno)
            name2_anno = name1_anno

        # Confirm names are same.
        if name1[:neg_index] != name2[:neg_index]:
            raise ParseException(
                    """Mismatched read names: {}, {}""".format(name1, name2))

        out1.write("{}\n{}\n+\n{}\n".format(name1_anno, seq1[clip_len1:],
            qual1[clip_len1:]))
        out2.write("{}\n{}\n+\n{}\n".format(name2_anno, seq2[clip_len2:],
            qual2[clip_len2:]))

def create_annotated_files_from_fastq(fp_extract_umis, in1, in2, out1, out2,
        has_index):
    """Using paired FASTQ files as input, create paired FASTQ output files with
    read name annotated with UMI.

    Args:
        fp_extract_umis (function): A function used to extract the UMIs.
        in1 (file): Read1 input fastq file.
        in2 (file): Read2 input fastq file.
        out1 (file): Read1 annotated output fastq file.
        out2 (file): Read2 annotated output fastq file.
        has_index (bool): Whether or not these FASTQ files have the index
            string (e.g. 'read_id 1:N:0:ATCACGTT')
    """

    # Walk the file and create a new annotated file
    while True:
        name1, seq1, junk, qual1 = \
                (in1.readline().rstrip() for i in range(4))
        name2, seq2, junk, qual2 = \
                (in2.readline().rstrip() for i in range(4))

        if not name1:
            # EOF
            break

        clip_len1, clip_len2, umi1, umi2 = fp_extract_umis(seq1, seq2)
        umis_anno = DELIM_ANNO_READ_PAIR.join([umi1, umi2])

        if has_index:
            name1, index_str1 = name1.split()
            name2, index_str2 = name2.split()

        if name1[-2:] == '/1':
            # Some datasets have a '/1' and '/2' at end of R1 and R2 read names
            # respectively.
            neg_index = -2
            name = name1[:-2]
            name_anno = "{}{}{}".format(name, DELIM_ANNO, umis_anno)
            record1 = "{}/1\n{}\n+\n{}\n".format(name_anno, seq1[clip_len1:],
                    qual1[clip_len1:])
            record2 = "{}/2\n{}\n+\n{}\n".format(name_anno, seq2[clip_len2:],
                    qual2[clip_len2:])
        elif has_index:
            # Some reads have the index included.
            neg_index = 0
            name = name1
            name_anno = "{}{}{}".format(name, DELIM_ANNO, umis_anno)
            record1 = "{} {}\n{}\n+\n{}\n".format(name_anno, index_str1,
                    seq1[clip_len1:], qual1[clip_len1:])
            record2 = "{} {}\n{}\n+\n{}\n".format(name_anno, index_str2,
                    seq2[clip_len2:], qual2[clip_len2:])
        else:
            # Some reads have neither the index nor the /1, /2
            neg_index = 0
            name = name1
            name_anno = "{}{}{}".format(name, DELIM_ANNO, umis_anno)
            record1 = "{}\n{}\n+\n{}\n".format(name_anno, seq1[clip_len1:],
                    qual1[clip_len1:])
            record2 = "{}\n{}\n+\n{}\n".format(name_anno, seq2[clip_len2:],
                    qual2[clip_len2:])

        # Confirm names are same.
        if name1[:neg_index] != name2[:neg_index]:
            raise ParseException(
                    """Mismatched read names: {}, {}""".format(name1, name2))

        out1.write(record1)
        out2.write(record2)

def create_annotated_file_from_fastq(fp_extract_umi, in1, out1, has_index):
    """Using single-end FASTQ file as input, create single-end FASTQ output
    file with read name annotated with UMI.

    Args:
        fp_extract_umis (function): A function used to extract the UMIs.
        in1 (file): Input fastq file.
        out1 (file): Annotated output fastq file.
        has_index (bool): Whether or not these FASTQ files have the index
            string (e.g. 'read_id 1:N:0:ATCACGTT')
    """

    # Walk the file and create a new annotated file
    while True:
        name1, seq1, junk, qual1 = \
                (in1.readline().rstrip() for i in range(4))

        if not name1:
            # EOF
            break

        clip_len1, umi1 = fp_extract_umi(seq1)
        umi_anno = umi1

        if has_index:
            name1, index_str1 = name1.split()

        if has_index:
            # Some reads have the index included.
            name = name1
            name_anno = "{}{}{}".format(name, DELIM_ANNO, umi_anno)
            record1 = "{} {}\n{}\n+\n{}\n".format(name_anno, index_str1,
                    seq1[clip_len1:], qual1[clip_len1:])
        else:
            # For reads that don't have an index.
            name = name1
            name_anno = "{}{}{}".format(name, DELIM_ANNO, umi_anno)
            record1 = "{}\n{}\n+\n{}\n".format(name_anno, seq1[clip_len1:],
                    qual1[clip_len1:])

        out1.write(record1)

def extract_single_umi_bioo(seq1):
    """Extract the Bioo UMI from a single end read.

    Args:
        seq1 (str): The full read1 (with UMI).
    Returns:
        (int, str): A 2-tuple, the int representing the amount
            to clip from the read, the str is the UMI of the read.
    """
    umi_len = len(UMIS_BIOO[0])
    # The Bioo kit uses a 1nt 'T' overhang for ligation.
    clip_len = umi_len + 1

    umi1 = seq1[:umi_len]

    return (clip_len, umi1)

def extract_paired_umis_bioo(seq1, seq2):
    """Extract the Bioo UMIs from a paired read.

    Args:
        seq1 (str): The full read1 (with UMI).
        seq2 (str): The full read2 (with UMI).
    Returns:
        (int, int, str, str): A 4-tuple, first two ints representing the amount
            to clip from the two reads respectively, second two strs are the
            UMIs of each read respectively.
    """
    umi_len = len(UMIS_BIOO[0])
    # The Bioo kit uses a 1nt 'T' overhang for ligation.
    clip_len = umi_len + 1
    umi1 = seq1[:umi_len]
    umi2 = seq2[:umi_len]

    return (clip_len, clip_len, umi1, umi2)

def parse_args(args):
    """Parse the command line arguments. """

    # Paired or single end? FASTQ or BAM?
    if args['<in2.fastq>']:
        # mode = FASTQ, PE
        paired = True
        filetype = 'FASTQ'
        # Convert ~ to real path
        args['<in1.fastq>'] = os.path.expanduser(args['<in1.fastq>'])
        args['<in2.fastq>'] = os.path.expanduser(args['<in2.fastq>'])
        input_files = [args['<in1.fastq>'], args['<in2.fastq>']]
    else:
        # Convert ~ to real path
        args['<input.fastq>'] = os.path.expanduser(args['<input.fastq>'])

        # Note: If you write the following for docopt:
        #   dupliganger remove-umi [options] <input.fastq>
        #   dupliganger remove-umi [options] <input.bam>
        # then it will always populate <input.fastq> and never populate
        # <input.bam>, hence the somewhat confusing names going on down
        # below...
        if is_bam(args['<input.fastq>']):
            # It is a bam, so hack docopt a bit
            args['<input.bam>'] = args['<input.fastq>']
            args['<input.fastq>'] = None
            paired = True if is_paired_bam(args['<input.bam>']) else False
            filetype = 'BAM'
            input_files = [args['<input.bam>']]
        else:
            # mode = FASTQ, SR
            paired = False
            filetype = 'FASTQ'
            input_files = [args['<input.fastq>']]

    # What fp_extract_umi to use?
    kit = args['--kit']
    if kit == KIT_BIOO:
        if paired:
            fp_extract_umi = extract_paired_umis_bioo
        else:
            fp_extract_umi = extract_single_umi_bioo
    else:
        raise CannotContinueException(
                """Kit {} is not supported.""".format(kit))

    # Figure out which function to use to write to output file.
    num_threads = int(args['--threads'])
    compress = args['--compress']
    if num_threads > 1 and which('pigz') and compress:
        # return a partial for pigzwrite
        # TODO: Maybe alter this to num_threads/2 if --paired-end?
        fp_write = functools.partial(pigzwrite, num_threads)
    elif which('gzip') and compress:
        # return a partial for gzwrite
        fp_write = gzwrite
    else:
        fp_write = functools.partial(open, mode = 'w')

    # Return an appropriate function pointer for annotation.
    if filetype == 'FASTQ':
        if paired:
            fp_anno = create_annotated_files_from_fastq
        else:
            fp_anno = create_annotated_file_from_fastq
    elif filetype == 'BAM':
        if paired:
            fp_anno = create_annotated_files_from_bam
        else:
            fp_anno = create_annotated_file_from_bam
    else:
        raise ControlFlowException(
                """ERR213: Not possible to be here.""")

    outdir = args_to_out_dir(args)

    if args['--force-paired'] and not paired:
        sys.stderr.write(
                "WARNING: Passed --force-paired but {} appears not to be "
                "paired-end.\n")
    force_paired = args['--force-paired']

    return (fp_extract_umi, fp_anno, fp_write, outdir, compress, force_paired,
            input_files)

def run(fp_extract_umi, fp_anno, fp_write, outdir, compress, force_paired,
        input_files):
    """Start the run.

    Args:
        fp_extract_umi (function): Function extract UMIs from single/paired
            reads, also returns the length(s) to clip.
        fp_anno (Function): Function to be used to parse the FASTQ or BAM file(s).
        fp_write (Function): Function to be used to write output file(s).
        outdir (str): Output directory for results
        compress (bool): Whether or not to compress the output.
        force_paired (bool): Whether user wants to force paired-end, even if
            if we didn't detect it.
        input_files ([str]): Array of input files to be parsed.
    """

    if fp_anno == create_annotated_files_from_bam or force_paired:
        # BAM, paired-end
        infile1 = input_files[0]
        out1, out2 = filename_in_bam_to_out_fqgz(infile1,
                SUFFIX_REMOVE_UMI, compress, True, outdir)
        out_files = [out1, out2]
        tmp_out1, tmp_out2 = tmpf_start(out1, out2)

        with bamopen(infile1) as in1, \
                fp_write(tmp_out1) as out1, \
                fp_write(tmp_out2) as out2:
            fp_anno(fp_extract_umi, in1, out1, out2)
        tmpf_finish(tmp_out1, tmp_out2)

    elif fp_anno == create_annotated_file_from_bam:
        # BAM, single-end
        infile1 = input_files[0]
        out1 = filename_in_bam_to_out_fqgz(infile1, SUFFIX_REMOVE_UMI,
                compress, False, outdir)[0]
        out_files = [out1]
        tmp_out1 = tmpf_start(out1)[0]

        with bamopen(infile1) as in1, \
                fp_write(tmp_out1) as out1:
            fp_anno(fp_extract_umi, in1, out1)
        tmpf_finish(tmp_out1)

    else:

        # Some datasets have look like this:
        #       @NS500451:139:H5TV5AFXX:1:11101:3928:1111 1:N:0:ATCACGTT
        # Others like this:
        #       @NS500451:139:H5TV5AFXX:1:11101:3928:1111
        # Detect.
        if is_gzipped(input_files[0]):
            with gzip.open(input_files[0], mode='rt') as in1:
                first_line = in1.readline()
        else:
            with io.open(input_files[0], mode='r', encoding='latin-1') as in1:
                first_line = in1.readline()
        has_index = True if len(first_line.split()) > 1 else False

        # FASTQ
        if len(input_files) == 2:
            # FASTQ, paired-end
            infile1, infile2 = input_files
            out1 = filename_in_to_out_fqgz(infile1, SUFFIX_REMOVE_UMI, compress,
                    outdir)
            out2 = filename_in_to_out_fqgz(infile2, SUFFIX_REMOVE_UMI, compress,
                    outdir)
            out_files = [out1, out2]
            tmp_out1, tmp_out2 = tmpf_start(out1, out2)

            with pgopen(1, infile1) as in1, \
                    pgopen(1, infile2) as in2, \
                    fp_write(tmp_out1) as out1, \
                    fp_write(tmp_out2) as out2:
                fp_anno(fp_extract_umi, in1, in2, out1, out2, has_index)
            tmpf_finish(tmp_out1, tmp_out2)

        elif len(input_files) == 1:
            # FASTQ, single-end
            infile1 = input_files[0]
            out1 = filename_in_to_out_fqgz(infile1, SUFFIX_REMOVE_UMI, compress,
                    outdir)
            out_files = [out1]
            tmp_out = tmpf_start(out1)[0]

            with pgopen(1, infile1) as in1, \
                    fp_write(tmp_out) as out1:
                fp_anno(fp_extract_umi, in1, out1, has_index)
            tmpf_finish(tmp_out)

        else:
            raise ControlFlowException(
                    """ERR911: Not possible to be here.""")

    return out_files

###############
### Classes ###
###############


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Quality trim reads.  This is a 'Trimmomatic' wrapper.

Usage:
    dupliganger qtrim [options] <input.fastq>
    dupliganger qtrim [options] <in1.fastq> <in2.fastq>


Note:
    Dupligänger supports (and autodetects) input FASTQ files that are gzipped.

Options:
    -h, --help
    -v, --verbose           Be verbose.
    -o OUT_DIR              Place results in directory OUT_DIR.
    --compress              Compress (gzip) output files.
    -t N, --threads N       Number of threads to run with [default: 1].
    -T S, --trimmomatic S   Pass 'S' as arguments to Trimmomatic. Don't forget to put quotes
                            around S if passing in more than one argument.
                            [default: 'LEADING:10 TRAILING:10 SLIDINGWINDOW:5:10 MINLEN:30']
    -p P, --phred P         Set P to 33 for phred33 encoded files, or 64 for
                            phred 64 encoded FASTQ files [default: 33].
"""

###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object

# Dupligänger imports
from dupliganger.exceptions import *
from dupliganger.constants import *
from dupliganger.common import (pgopen, tmpf_start, tmpf_finish,
        filename_in_to_out_fqgz, pe_log_filename, se_log_filename,
        args_to_out_dir)

## Other imports
from docopt import docopt

# For converting ~ to full path
import os

# For external command execution
import subprocess, shlex


#################
### Constants ###
#################

FIXME_TRIMMOMATIC_SRC = '~/src/Trimmomatic-0.36/trimmomatic-0.36.jar'

#################
### Functions ###
#################

def qtrim(num_threads, phred, trimmomatic_args, *input_output_files):
    """Quality trim reads with Trimmomatic.

    Args:
        num_threads (int): How many threads to run with.
        phred (int): Phred encoding of fastq file(s). (33 or 64)
        trimmomatic_args (str): Arguments (in string format) to pass to
            Trimmomatic (e.g. TRAILING:10 MINLEN:50).
        input_output_files ([str]): Input and output files.
    """

    threads = "-threads {}".format(num_threads)
    unpaired1 = ''

    if len(input_output_files) == 8:
        in1, in2, out1, out2, unpaired1, unpaired2, out_log, trim_log = \
                input_output_files
        ins = "{} {}".format(in1, in2)
        outs = "{} {} {} {}".format(out1, unpaired1, out2, unpaired2)
        mode = 'PE'
    elif len(input_output_files) == 4:
        in1, out1, out_log, trim_log = input_output_files
        ins = "{}".format(in1)
        outs = "{} {}".format(out1, unpaired1)
        mode = 'SE'
    else:
        raise ControlFlowException("""ERR911: Not possible to be here.""")

    phred = "-phred{}".format(phred)
    trimlog_str = "-trimlog {}".format(trim_log)

    cmd = "java -jar {} {} {} {} {} {} {} {}".format(
            os.path.expanduser(FIXME_TRIMMOMATIC_SRC), mode, threads, phred,
            trimlog_str, ins, outs, trimmomatic_args)

    print("Running:\n\t{}".format(cmd))
    cmd = shlex.split(cmd)

    with open(out_log, 'w') as f:
        p = subprocess.Popen(cmd, stdout=f, stderr=subprocess.STDOUT, bufsize=-1)
        retval = p.wait()
    if retval > 0:
        raise ExecutionException(
                """Failure to execute properly. See {}""".format(out_log))

def parse_args(args):
    """Parse the command line arguments.
    """

    # Convert ~ to real path, and get input files
    if args['<in2.fastq>']:
        fin1 = os.path.expanduser(args['<in1.fastq>'])
        fin2 = os.path.expanduser(args['<in2.fastq>'])
        input_files = [fin1, fin2]
    else:
        fin = os.path.expanduser(args['<input.fastq>'])
        input_files = [fin]

    num_threads = int(args['--threads'])
    compress = args['--compress']
    outdir = args_to_out_dir(args)

    phred = args['--phred']
    if phred not in ('33', '64'):
        raise CannotContinueException(
                """Phred score {} is not supported.""".format(phred))
    phred = int(phred)

    trimmomatic_args = args['--trimmomatic'].strip("'").strip('"').strip("'")

    return (num_threads, outdir, compress, phred, trimmomatic_args, input_files)

def run(num_threads, outdir, compress, phred, trimmomatic_args, input_files):
    """Start the run.

    Args:
        num_threads (int): How many threads to run with.
        outdir (str): Output directory for results
        compress (bool): Whether or not to compress the output.
        phred (int): Phred encoding of fastq file(s). (33 or 64)
        trimmomatic_args (str): Arguments (in string format) to pass to
            Trimmomatic (e.g. TRAILING:10 MINLEN:50).
        input_files ([str]): Input files.
    """

    if len(input_files) == 2:
        in1, in2 = input_files
        out1 = filename_in_to_out_fqgz(in1, SUFFIX_QTRIM, compress, outdir)
        out2 = filename_in_to_out_fqgz(in2, SUFFIX_QTRIM, compress, outdir)
        out_files = [out1, out2]
        out_log = pe_log_filename(SUFFIX_QTRIM, out2)
        trim_log = pe_log_filename(SUFFIX_QTRIM, out2, 'trimlog')
        # i.e. for those that end up unpaired because partner trimmed too much
        unpaired_out1 = filename_in_to_out_fqgz(out1, SUFFIX_QTRIM_UNPAIRED,
                compress, outdir)
        unpaired_out2 = filename_in_to_out_fqgz(out2, SUFFIX_QTRIM_UNPAIRED,
                compress, outdir)

        tmp_out1, tmp_out2, tmp_unpaired_out1, tmp_unpaired_out2 = tmpf_start(out1,
                out2, unpaired_out1, unpaired_out2)

        qtrim(num_threads, phred, trimmomatic_args, in1, in2, tmp_out1,
                tmp_out2, tmp_unpaired_out1, tmp_unpaired_out2, out_log,
                trim_log)
        tmpf_finish(tmp_out1, tmp_out2, tmp_unpaired_out1, tmp_unpaired_out2)

    elif len(input_files) == 1:
        in1 = input_files[0]
        out1 = filename_in_to_out_fqgz(in1, SUFFIX_QTRIM, compress, outdir)
        out_files = [out1]
        out_log = se_log_filename(SUFFIX_QTRIM, out1)
        trim_log = se_log_filename(SUFFIX_QTRIM, out1, 'trimlog')

        tmp_out1 = tmpf_start(out1)[0]

        qtrim(num_threads, phred, trimmomatic_args, in1, tmp_out1, out_log,
                trim_log)
        tmpf_finish(tmp_out1)

    else:
        raise ControlFlowException("""ERR911: Not possible to be here.""")

    return out_files


###############
### Classes ###
###############


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

"""Remove PCR duplicates from an alignment file (SAM format).

Usage:
    dupliganger dedup [options] <alignment-file>


Output (subject to change):
    1) <alignment-file>.dups_removed.sam - SAM file with PCR duplicates removed
        from <alignment-file>.
    2) <alignment-file>.dups_flagged.sam - SAM file with PCR duplicates flagged
        and retained in <alignment-file>.  This file is not generated by
        default.
    3) <alignment-file>.duplicates.sam - SAM file of reads called PCR
        duplicates.
    4) <alignment-file>.dup_groups.samlike - SAM-like file that contains groups
        of reads (or read-pairs) that have the same location and UMIs.  Mostly
        useful for sanity checking.
    5) <alignment-file>.umi_errors.sam - SAM file that contains reads rejected
        due to errors in one or both of their UMIs.  Only applicable if UMIs
        are known a priori (e.g. when --kit is set to bioo).

Options:

    -h, --help              Print this help.
    -v, --verbose           Be verbose
    -o OUT_DIR              Place results in directory OUT_DIR. Defaults to the
                            current directory.
    -u, --unpaired          Reads are unpaired [default: False].
    -k KIT, --kit KIT       (Optional) The kit used (case insensitive). If
                            'bioo' is given, then ...
    --store STORE           Storage backend: 'lmdb' or 'memory' [default:
                            memory].
    --random-seed SEED      For those wanting complete control, change the
                            random seed from its normally hardcoded value.

Options when UMIs are known a priori (e.g. --kit=Bioo):

    -K, --keep-bad-umis     Default is to remove a read (or a read-pair) if
                            there is an error in the read's UMI (or one of the
                            read-pair's UMIs).  Use this flag to retain reads
                            that have an error in their UMIs. Only applicable
                            if --kit bioo.  See also --correct-umis.
    -c, --correct-umis      Perform UMI correction if an UMI is at most 1nt
                            Hamming distance away from at most one known UMI.
                            Only applicable if --kit bioo.  See also
                            --keep-bad-umis. NOT CURRENTLY IMPLEMENTED.

Output Options:

    --no-write-dedupped-sam  Do not write the default output of dupliganger
                             (which is a SAM file with PCR duplicates removed).
                             See also --write-flagged-sam.
    --write-flagged-sam      Write out a SAM file with PCR duplicates included
                             but flagged as PCR duplicates (SAM flag 0x400). See
                             also --no-write-dedupped-sam.
    --no-write-dup-sam       Do no write a SAM file that contains only
                             PCR duplicates.  See also --write-dup-group-file.
    --no-write-dup-group-file  Write a SAM-like file that contains groups of
                               reads (or read-pairs) that have the same location
                               and UMIs.
    --no-write-sam-headers   Do not write SAM headers.

    --unannotate-read-name   Remove dupliganger read-name annotations from
                             final output.  NOT IMPLEMENTED.

Debug Options:

    --debug-no-build-read-and-loc-dbs   For debugging: Don't build the read and
                                        location DBs.

    --debug-switch S        Fun debug switch [default: nothing].
    --dump-rg-db            Dump the ReadGroup database (for debugging purposes only).
    --dump-loc-db           Dump the Location database (for debugging purposes only).
    --dump-dup-group-db     Dump the Duplication Group database (for debugging purposes only).
    --dump-dup-db           Dump the Duplication database (for debugging purposes only).
    --dump-umi-error-db     Dump the UMI error database (for debugging purposes only).

"""

###############
### Imports ###
###############

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object
# py-lmdb uses bytes() for py3 and str() for py2.
# This package has different code for py2 and py3.
# And importing that future 'object' has a bug that screws up __slots__ in
# py2 (causes different behavior than in py3).

# For iteration
from future.utils import iteritems, itervalues

# For range as iterator in py23
from builtins import range

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *
from dupliganger.common import (setup_report_db, args_to_out_dir, sambamopen,
        filename_in_to_out_sambam, tmpf_start, tmpf_finish, memory_info)
from dupliganger.db import (ParentDbDict, ParentDbLmdb, SimpleBucketDict,
        SimpleBucketLmdb, SimpleObjectDbDict, SimpleObjectDbLmdb)
from dupliganger.sam import (Read, ReadGroup, to_location_key_with_5p_trimming)
from dupliganger.build_read_and_loc_dbs import write_to_read_and_location_dbs

# debug
from dupliganger.db import LocationBucketDb

## Other imports
from docopt import docopt

# For filename fixing
import os

# For debug timing
import time

# For some printing to stdout
import sys

# For Bioo UMI correction
import distance

# For function partials
import functools

# For choosing by lowest read name the winner/losers
from operator import attrgetter

# For choosing winner of a DupGroup
import random

#################
### Constants ###
#################


#################
### Functions ###
#################

def choose_winner_and_losers_sorted_read_name(fp_get_read_group, dup_group):
    """Given a DupGroup (i.e. a list of read_group_ids), chooses as the winner
    the first by alphanumeric sort.

    Args:
        fp_get_read_group ((Transaction, str) -> ReadGroup): Function pointer
            that retrieves a ReadGroup from the read_group_db when passed a
            read_group_id.
        dup_group (DupGroup): A DupGroup filled up with ReadGroups.

    Returns:
        (str, (str,)): A two tuple: First member is the winning read_group_id,
            second member is a list of the losing read_group_ids.
    """

    dup_group = (fp_get_read_group(rg_id) for rg_id in dup_group)
    dup_group = sorted(dup_group, key=attrgetter('name'))
    winner = dup_group[0]
    losers = dup_group[1:]
    return (winner, losers)

def choose_winner_and_losers_longest_highest_quality(fp_get_read_group, dup_group):
    # TODO: Question, catherine, which to prioritize? Length or quality?

    # Note: To do this would cost more.  We're not currently storing average
    # quality score or the QUAL field in the read_group_db, so presumably we'd
    # have to calculate this upon reading the input SAM file, and store it
    # (without messing up the currently repr/load semantics). (Nor the length
    # of the sequence, but that's available via parse_cigar). Certainly
    # do-able, but I'm happy with the lack of bias provided by
    # choose_winner_losers_random_fixed_seed.
    pass

def choose_winner_and_losers_random_fixed_seed(fp_get_read_group, dup_group):
    """Given a DupGroup (i.e. a list of read_group_ids), chooses a random
    winner.  The random seed is fixed at the beginning of dedup.

    Args:
        fp_get_read_group ((Transaction, str) -> ReadGroup): Function pointer
            that retrieves a ReadGroup from the read_group_db when passed a
            read_group_id.
        dup_group ([str]): A list of read_group_ids representing from one
            DupGroup.

    Returns:
        (str, (str,)): A two tuple: First member is the winning read_group_id,
            second member is a list of the losing read_group_ids.
    """
    ###dup_group = sorted((fp_get_read_group(rg_id) for rg_id in dup_group))

    # Sort for determinism
    dup_group = sorted(dup_group)
    # Convert dup_group to a list of its member ReadGroups.
    dup_group = [fp_get_read_group(rg_id) for rg_id in dup_group]

    # Choose a winner randomly
    winner_idx = random.randrange(0, len(dup_group))
    winner = dup_group.pop(winner_idx)
    losers = dup_group
    return (winner, losers)

def log_winner_losers(winner, losers):
    """Logs winner and losers, somehow.  For now, just printing...
    """
    pass

def umi_bioo_report(seq_umi):
    """Generates a 'report' about a given sequenced UMI. It finds the Bioo UMIs
    closest in Hamming distance to the sequenced_umi.  It reports that Hamming
    distance along with all of the Bioo UMIs that distance away from the
    sequenced_umi.

    Args:
        seq_umi (str): The uncorrected UMI from the read sequence.
    Returns:
        (int, [str]): The report, a two-tuple.  First member of the tuple is
            the Hamming distance, second member is all the Bioo UMIs that
            distance away.
    """
    dist = inf = 999
    potential_umis = []

    if seq_umi in UMIS_BIOO:
        potential_umis = [seq_umi]
        dist = 0
    else:
        potential_umis = []
        for curr_dist in range(1,9):
            for umi in UMIS_BIOO:
                d = distance.hamming(umi, seq_umi)
                if d == curr_dist:
                    potential_umis.append(umi)
                    dist = curr_dist
            if dist < inf:
                # Found at least one
                break

    # if dist == inf:
    #     # TODO: Remove this code if not found after this...
    #     # Hrm, fixed the bug above (range(1,9) was previously range(1,8)) so
    #     # being here should be impossible.  Announce anomoly.
    #     msg = "WARNING umi_bio_report: dist == infinity, seq_umi = {}".format(
    #                 seq_umi)
    #     sys.stderr.write(msg)

    return (dist, potential_umis)

def process_location_pe_randomer_1nt(report_db, read_groups):
    """Does many things:

        * Partitions ReadGroups at this location into DupGroups (one DupGroup
          per umi).
        * Updates report_db for several metrics (what catherine wanted, and
            more):
                - levels of duplications (how many found)
                - number of duplicates we would have thrown out if not for UMIs.
                - more not listed here.

    This is the paired-end randomer version of the family of process_location()
    functions.

    Note: This function does *not* care whether there are one or two UMIs.  It
    simply uses the entire UMI string.  For example, if the read name were
    'D00597:180:C7NMDANXX:6:1101:1184:39633-GGCCTAAT^AGCTCTAG;2^0', then the
    umi_str would be 'GGCCTAAT^AGCTCTAG'.

    Args:
        report_db (dict): A database for metrics to report.
        read_groups (dict(str) -> [ReadGroup]): Dictionary, key: read_group_id,
            val: list of read_groups.
    Returns:
        [[str]]: A list of lists of ReadGroup ids.  Each inner list of
            ReadGroup ids constitutes a DupGroup.

    """
    # key: umi_str, val: list of read_group_ids with that umi_str
    rg_ids_by_umi_str = {}

    for read_group_id, read_group in iteritems(read_groups):

        # Note: This won't support multi-mapped alignments like other parts of
        # the code do...

        ## Pull out the UMIs and examine them.

        # D00597:180:C7NMDANXX:6:1101:1184:39633-GGCCTAAT^AGCTCTAG;2^0
        anno = read_group[0].qname.split(DELIM_ANNO)[1]
        # GGCCTAAT^AGCTCTAG;2^0
        umi_str = anno.split(DELIM_ANNO_TYPE)[0]
        # GGCCTAAT^AGCTCTAG

        rg_ids_by_umi_str.setdefault(umi_str, [])
        rg_ids_by_umi_str[umi_str].append(read_group_id)

    ## Finalize DupGroups
    dup_groups = rg_ids_by_umi_str.values()

    # report_db: update number of unique location/umi_str combinations
    # There is another place in write_to_dup_group_db where this metric is
    # incrememted.
    report_db[LOG_NUM_UNIQUE_UMI_AND_LOCATION_COMBINATIONS] += len(dup_groups)

    # Only DupGroups with more than one member are real DupGroups.
    dup_groups_larger_than_one_member = [ dg for dg in dup_groups if len(dg) > 1 ]

    # report_db: log number of true DupGroups:
    report_db[LOG_NUM_DUP_GROUPS] += len(dup_groups_larger_than_one_member)

    return dup_groups_larger_than_one_member

def process_location_pe_bioo_1nt(report_db, umi_error_db, reject_umi_errors,
        correct_umis, read_groups):
    """Does many things:

        * Partitions ReadGroups at this location into DupGroups (one DupGroup
          per umi-pair).
        * Identifies ReadGroups which have one or more errors in their UMIs,
          and writes those to the umi_error_db (eventually these are used to
          write out dupliganger SAM TAGs, just for reads with UMI errors).
        * Optionally conservatively corrects UMIs with errors (1nt away from
          exactly one known UMI) (Actuall, not implemented at this time!)
        * Updates report_db for several metrics (what catherine wanted, and
            more):
                - levels of duplications (how many found)
                - number of duplicates we would have thrown out if not for UMIs.
                - number of ReadGroups rejected due to errors in UMIs.
                - more not listed here.

    This is the paired-end Bioo version of the family of process_location()
    functions.

    Args:
        report_db (dict): A database for metrics to report.
        reject_umi_errors (bool): Whether or not to reject the umis with errors.
        umi_error_db (dict): DB of read groups with one or more UMI errors.
        correct_umis (bool): Whether or not to correct the umis.
        read_groups (dict(str) -> [ReadGroup]): Dictionary, key: read_group_id,
            val: list of read_groups.
    Returns:
        [[str]]: A list of lists of ReadGroup ids.  Each inner list of
            ReadGroup ids constitutes a DupGroup.

    Dupligänger SAM TAGS

        'sam_tags' is a tab separated list of dupliganger-defined SAM TAGs. It's
        in flux right now.  Here's the current specification.

        Dupligänger SAM file TAGs specification:

            NOTE! SUBJECT TO CHANGE!!!

            d1:i:count      Read1's UMI is 'count' nts hamming distance from a
                            known UMI.
            d2:i:count      Same as d1 but for Read2.
            n1:i:count      Read1's UMI is closest to 'count' known UMIs.
            n2:i:count      Same as n1, but for Read2.
            c1:Z:nts        There is only one case where a sequenced UMI might get
                            corrected (1nt away from just one known UMI).  IF we
                            end up doing this type of correction, then the
                            corrected UMI used will be 'nts', and only the read
                            that got corrected will get this TAG.
                            (Note: Uncorrected UMI is in the annotated read name.)
            c2:Z:nts        Same as c1, but for Read2.
    """
    # key: umi_pair, val: list of read_group_ids with that umi_pair.
    rg_ids_by_umi_pair = {}

    for read_group_id, read_group in iteritems(read_groups):

        # Note: This won't support multi-mapped alignments like other parts of
        # the code do...

        ## Pull out the UMIs and examine them.

        # D00597:180:C7NMDANXX:6:1101:1184:39633-GGCCTAAT^AGCTCTAG;2^0
        anno = read_group[0].qname.split(DELIM_ANNO)[1]
        # GGCCTAAT^AGCTCTAG;2^0
        umis = anno.split(DELIM_ANNO_TYPE)[0]
        # GGCCTAAT^AGCTCTAG
        umi1, umi2 = umis.split(DELIM_ANNO_READ_PAIR)
        # [GGCCTAAT, AGCTCTAG]
        dist1, potential_umis1 = umi_bioo_report(umi1)
        dist2, potential_umis2 = umi_bioo_report(umi2)
        len_potential_umis1 = len(potential_umis1)
        len_potential_umis2 = len(potential_umis2)

        ## Add dupliganger SAM TAGs and optionally correct UMIs.
        if dist1 >= 1 or dist2 >= 1:
            # Found UMI error!

            sam_tags1 = "{}{}\t{}{}".format(
                     SAM_TAG_READ1_HAMMING_DIST_UMI, dist1,
                     SAM_TAG_READ1_NUM_CLOSEST_UMIS, len_potential_umis1)
            sam_tags2 = "{}{}\t{}{}".format(
                     SAM_TAG_READ2_HAMMING_DIST_UMI, dist2,
                     SAM_TAG_READ2_NUM_CLOSEST_UMIS, len_potential_umis2)
            # Could we have corrected?
            if (    dist1 <= 1 and len_potential_umis1 <= 1 and
                    dist2 <= 1 and len_potential_umis2 <= 1):
                # Yes, we could have. Annotate.
                if dist1 == 1 and len_potential_umis1 == 1:
                    sam_tags1 += '\t{}{}'.format(SAM_TAG_READ1_CLOSEST_UMI,
                            potential_umis1[0])
                    if correct_umis:
                        # CORRECT UMI1!
                        umi1 = potential_umis1[0]
                if dist2 == 1 and len_potential_umis2 == 1:
                    sam_tags2 += '\t{}{}'.format(SAM_TAG_READ2_CLOSEST_UMI,
                            potential_umis2[0])
                    if correct_umis:
                        # CORRECT UMI2!
                        umi2 = potential_umis2[0]

            # Write to the umi_error_db:
            # NOTE: Assumption: The umi_error_db will be small, so just using a
            # simple dict here...
            umi_error_db[read_group.name] = (sam_tags1, sam_tags2)

            # report_db: count of UMI errors (largest dist only reported)
            dist = dist1 if dist1 > dist2 else dist2
            metric = "{}{}".format(LOG_NUM_READ_GROUPS_WITH_UMI_ERROR_DIST,
                    dist)
            report_db[metric] += 1
            # report_db: total number of read_groups with UMI errors
            report_db[LOG_NUM_READ_GROUPS_WITH_UMI_ERROR] += 1


        if reject_umi_errors and (dist1 >= 1 or dist2 >= 1):
            ## User requested rejection of ReadGroups with UMI errors.

            # report_db: count of rejects due to UMI errors
            dist = dist1 if dist1 > dist2 else dist2
            metric = "{}{}".format(LOG_NUM_READ_GROUPS_REJECTED_DUE_TO_UMI_ERROR_DIST,
                    dist)
            report_db[metric] += 1

        else:
            ## User requested keep all ReadGroups (even those with UMI errors).

            umi_pair = ",".join([umi1, umi2])
            rg_ids_by_umi_pair.setdefault(umi_pair, [])
            rg_ids_by_umi_pair[umi_pair].append(read_group_id)

    ## Finalize DupGroups
    dup_groups = rg_ids_by_umi_pair.values()

    # report_db: update number of unique location/umi_pair combinations
    # There is another place in write_to_dup_group_db where this metric is
    # incrememted.
    report_db[LOG_NUM_UNIQUE_UMI_AND_LOCATION_COMBINATIONS] += len(dup_groups)

    # Only DupGroups with more than one member are real DupGroups.
    dup_groups_larger_than_one_member = [ dg for dg in dup_groups if len(dg) > 1 ]

    # report_db: log number of true DupGroups:
    report_db[LOG_NUM_DUP_GROUPS] += len(dup_groups_larger_than_one_member)

    return dup_groups_larger_than_one_member

def put_dup_groups(dup_group_db, dup_groups):
    """For each duplicate group, given a duplicate group (in the form of a list
    of read_group_ids), write that duplicate group to the DupGroupDB, using
    either an existing DupGroup or creating a new one.

    Args:
        dup_group_db (DB): The DupGroup database.
        dup_groups ([[str]]): A list of lists of ReadGroup ids.  Each list of
            ReadGroup ids is a "DupGroup" (i.e. a Duplicate Group).

    The end result might look something like this:
        For these two dup groups:
            (1,2,3) and (4,5)
        You would end up with two total DupGroup objects, in the
        DupGroupDB (dgdb) like so:
            dgdb[1] -> DupGroupObjectA
            dgdb[2] -> DupGroupObjectA
            dgdb[3] -> DupGroupObjectA
            dgdb[4] -> DupGroupObjectB
            dgdb[5] -> DupGroupObjectB
    """

    for dup_group in dup_groups:

        # Is there any existing DupGroup for any of our ReadGroups?
        existing = set([ dup_group_db[rg_id] for rg_id in dup_group if rg_id in
                dup_group_db ])
        if len(existing) > 1:
            raise(CannotContinueException,
                    "Found more than one DupGroup representing: {}".format(
                        existing))
        elif len(existing) == 1:
            # Found an existing DupGroup, use that one
            dup_group_obj = existing[0]
        else:
            dup_group_obj = DupGroup()

        for read_group_id in dup_group:
            dup_group_obj.add(read_group_id)
            dup_group_db[read_group_id] = dup_group_obj

def write_to_dup_group_db(report_db, parent_db, read_group_db, loc_db,
        dup_group_db, records_per_txn, fp_process_location):
    """
    Arguments:
        fp_process_location (function): Function that does a lot of stuff, but
            mainly, processes a given location, partitioning ReadGroups at that
            location into DupGroups (by umi-pair), optionally rejects and/or
            corrects UMIs, logs things, etc.
    """

    # Populate the DupGroupDB
    with parent_db.begin(True) as txn_loc:

        fp_get_read_group = read_group_db.fp_get(txn_loc)
        # fp_put_read_group = read_group_db.fp_put(txn_loc)

        for location_key, read_group_ids in loc_db.iteritems(txn_loc):
            # record another location
            report_db[LOG_NUM_LOCATIONS] += 1
            if len(read_group_ids) > 1:
                # Found a location with more than one ReadGroup, potential
                # duplicate. Build a local dict of just the read_groups found
                # at this location
                read_groups = {}
                for read_group_id in read_group_ids:
                    read_group = fp_get_read_group(read_group_id)
                    read_groups[read_group_id] = read_group
                # Identify DupGroups and Rejected ReadGroups (rejected due to
                # errors in UMIs)
                dup_groups = fp_process_location(read_groups)
                # Write to the DupGroupDb
                put_dup_groups(dup_group_db, dup_groups)
            else:
                # Only one read group at this location, therefore increment the
                # number of unique location/umi_pair combos.  (When there are
                # more than one read_groups per location, this metric is
                # updated in process_location() (see a few lines above))
                metric = LOG_NUM_UNIQUE_UMI_AND_LOCATION_COMBINATIONS
                report_db[metric] += 1

def write_to_dup_db(report_db, parent_db, read_group_db, dup_group_db, dup_db,
        records_per_txn, fp_choose_winner_and_losers, fp_log_winner_losers):
    """
    Args:
        fp_choose_winner_and_losers (function): Function that determines
            which ReadGroup of a DupGroup is kept.
    """

    # NOTE! report_db not currently being used here?

    first_loop = True
    done = False
    time1 = None
    time2 = None
    time_in_write_db = 0

    # Grab a unique list of DupGroups
    # TODO: potentially expensive? measure...
    dg_ids = {}
    dup_groups = []
    for dup_group in itervalues(dup_group_db):
        if id(dup_group) not in dg_ids:
            dg_ids[id(dup_group)] = True
            dup_groups.append(dup_group)
    dup_groups = sorted(dup_groups)

    # Create an items iterator for the DupGroups
    iter_dup_group = iter(dup_groups)

    # Populate the DupDB
    while (not done):
        if first_loop:
            first_loop = False
        else:
            time2 = time.time()
            time_in_write_db += (time2 - time1)
            print("Commit dup_group_db at record_count = {}, time taken = {}, current mem MBs: {}".format(
                record_count, time2 - time1, memory_info(True)))
        time1 = time.time()
        with parent_db.begin(True) as txn_dup:
            # sys.stdout.write("Begin commit to DupDB... ")
            # functions
            fp_get_read_group = read_group_db.fp_get(txn_dup)
            fp_put_dup_db = dup_db.fp_put(txn_dup)
            fp_choose_winner_and_losers2 = functools.partial(
                    fp_choose_winner_and_losers, fp_get_read_group)
            (done, record_count) = write_to_dup_db_txn(iter_dup_group,
                    records_per_txn, fp_put_dup_db,
                    fp_choose_winner_and_losers2, fp_log_winner_losers)

def write_to_dup_db_txn(iter_dup_group, records_per_txn, fp_put_dup_db,
        fp_choose_winner_and_losers, fp_log_winner_losers):
    """Write a 'records_per_txn' number of to DB.

    Args:
        iter_dup_group (variable): An items iterator for the set of DupGroups.
        records_per_txn (int): The number of records.
        fp_choose_winner_and_losers (function): Function to ...
    Returns:
        (bool, int): 2-tuple with elements a) done or not, b) current
            record count.
    """
    record_count = 0
    curr_dup_group = None
    done = False

    try:
        while record_count < records_per_txn:
            curr_dup_group = next(iter_dup_group)
            winner, losers = fp_choose_winner_and_losers(curr_dup_group)
            for loser in losers:
                fp_put_dup_db(loser.name, [''])
            fp_log_winner_losers(winner, losers)
            record_count += 1
    except(StopIteration):
        done = True
    return(done, record_count)

def write_dup_group_sam_like_file(parent_db, read_group_db, dup_group_db,
        dup_group_sam_like):
    """Writes a "DupGroup" SAM-like file - Similar to a SAM file; the
    ReadGroups from a DupGroup are printed together; each DupGroup is separated
    from other DupGroups by new lines."""

    ## Write only the dup_group_sam_like file.
    with parent_db.begin(False) as txn, \
            open(dup_group_sam_like, 'w') as f_dup_group_sam_like:

        fp_get_read_group = read_group_db.fp_get(txn)

        # TODO: Potentially expensive?  And you've already generated this
        # earlier...
        # Grab a unique list of DupGroups
        time1 = time.time()
        dg_ids = {}
        dup_groups = []
        for dup_group in sorted(itervalues(dup_group_db)):
            if id(dup_group) not in dg_ids:
                dg_ids[id(dup_group)] = True
                dup_groups.append(dup_group)
        time2 = time.time()
        print('DEBUG TIMING A: It took {}s to generate list of unique dup groups.'.format(
                time2 - time1))

        for dup_group in dup_groups:
            for read_group_id in dup_group:
                read_group = fp_get_read_group(read_group_id)
                for read in read_group:
                    f_dup_group_sam_like.write(repr(read))
                    f_dup_group_sam_like.write('\n')
            # This newline separates each dup_group from the next
            f_dup_group_sam_like.write('\n\n')

def write_output_files_pe(parent_db, read_group_db, dup_db, umi_error_db,
        input_file, reject_umi_errors, dedupped_sam, flagged_sam, dup_only_sam,
        rejects_sam, write_dedupped_sam, write_flagged_sam, write_dup_only_sam,
        write_dup_group_sam_like, write_umi_error_rejects, write_sam_headers):
    """Writes the output files for paired-end input. Including:
        * A dedupped SAM file - no PCR duplicates included
        * A flagged SAM file - Identical to the input_file, but with PCR
            duplicates flagged.
        * A dup-only SAM file - Only contains the PCR duplicates.
        * A umi-error-rejects file. - Reads rejected due to error in UMI.
    """
    ## Write everything except the dup_group_sam_like file.
    with parent_db.begin(False) as txn, \
            sambamopen(input_file) as fin, \
            open(dedupped_sam, 'w') as f_dedupped_sam, \
            open(flagged_sam, 'w') as f_flagged_sam, \
            open(dup_only_sam, 'w') as f_dup_only_sam, \
            open(rejects_sam, 'w') as f_umi_reject_sam:

        # Setup
        fp_get_dup = dup_db.fp_get(txn)
        aln_lines = []
        group_count = 0
        total_count = 0
        remaining_lines = True
        aln_line = None
        prev_qname = None

        # Check for @HD line.
        line = fin.readline()
        if line[0:3] == '@HD':
            if write_sam_headers:
                # Found '@HD' line, which must be the first line. Write it.
                f_dedupped_sam.write(line)
                f_flagged_sam.write(line)
                f_dup_only_sam.write(line)
                f_umi_reject_sam.write(line)
        else:
            # No @HD line, start at beginning
            fin.seek(0)

        # Safe to write @PG line now.
        if write_sam_headers:
            pg_line = '@PG\tID:{}\tPN:{}\tVN:v{}\tCL:{}\n'.format(
                    'dupliganger', 'dupliganger', __version__,
                    ' '.join(sys.argv))
            f_dedupped_sam.write(pg_line)
            f_flagged_sam.write(pg_line)
            f_dup_only_sam.write(pg_line)
            f_umi_reject_sam.write(pg_line)

        # Write out headers
        while True:
            aln_line = fin.readline()
            if aln_line[0] == '@':
                if write_sam_headers:
                    f_dedupped_sam.write(aln_line)
                    f_flagged_sam.write(aln_line)
                    f_dup_only_sam.write(aln_line)
                    f_umi_reject_sam.write(aln_line)
            else:
                prev_qname, _ = aln_line.split('\t', 1)
                aln_lines.append(aln_line)
                break

        ## Write out alignment lines
        while remaining_lines:
            # Outer loop: Just keep going until EOF.
            while True:
                # Inner loop: Collect all alignments with same QNAME.
                total_count += 1
                aln_line = fin.readline()
                if not aln_line:
                    # EOF
                    qname = None
                    remaining_lines = False
                    break
                qname, _ = aln_line.split('\t', 1)

                if prev_qname and prev_qname != qname:
                    # New qname found, write out!
                    break
                else:
                    aln_lines.append(aln_line)
                    prev_qname = qname

            ## Have a group of alignment lines with same QNAME, write to files.

            # First, add dupliganger SAM TAGs if its in the umi_error_db
            if prev_qname in umi_error_db:
                # Error in UMI, add sdd-specific umi error tags.
                new_aln_lines = []
                for i, line in enumerate(aln_lines):
                    # Update the alignment line with sam tags
                    line_with_sam_tags = '\t'.join(
                            (aln_lines[i].rstrip(), umi_error_db[prev_qname][i]))
                    line_with_sam_tags += '\n'
                    new_aln_lines.append(line_with_sam_tags)
                aln_lines = new_aln_lines

            # Now write files
            if fp_get_dup(prev_qname):
                # Found a duplicate.
                # Update flag field to indicate its a dup.
                # Write updated line to dup_only and flagged_sam files.
                for line in aln_lines:
                    # recall FLAG is the second field in sam files
                    parts = line.split('\t', 2)
                    parts[1] = str(int(parts[1]) | SAM_FORMAT_FLAG_DUPLICATE)
                    flagged_line = '\t'.join(parts)
                    f_flagged_sam.write(flagged_line)
                    f_dup_only_sam.write(flagged_line)
            elif prev_qname in umi_error_db and reject_umi_errors:
                # Error in UMI and no rejects wanted by user: write to umi_reject file
                for line in aln_lines:
                    f_umi_reject_sam.write(line)
            else:
                # No duplicate: write to dedupped and flagged files
                for line in aln_lines:
                    f_dedupped_sam.write(line)
                    f_flagged_sam.write(line)

            # Prep for next round
            prev_qname = qname
            aln_lines = [aln_line]
            group_count += 1

def parse_args(args):
    """Parse the command line arguments."""

    debug_switch = args['--debug-switch']
    dump_rg_db = args['--dump-rg-db']
    dump_loc_db = args['--dump-loc-db']
    dump_dup_group_db = args['--dump-dup-group-db']
    dump_dup_db = args['--dump-dup-db']
    dump_umi_error_db = args['--dump-umi-error-db']

    random_seed = args['--random-seed']

    write_dedupped_sam = not args['--no-write-dedupped-sam']
    write_flagged_sam = args['--write-flagged-sam']
    write_dup_only_sam = not args['--no-write-dup-sam']
    write_dup_group_sam_like = not args['--no-write-dup-group-file']
    write_sam_headers = not args['--no-write-sam-headers']

    paired = False if args['--unpaired'] else True
    reject_umi_errors = not args['--keep-bad-umis']
    correct_umis = args['--correct-umis']
    build_read_and_loc_dbs = not args['--debug-no-build-read-and-loc-dbs']

    # Convert ~ to real path
    input_file = os.path.expanduser(args['<alignment-file>'])

    # Which kit?
    kit = str(args['--kit']).lower()
    if kit == KIT_BIOO:
        write_umi_error_rejects = True
    else:
        write_umi_error_rejects = False

    if correct_umis and kit != KIT_BIOO:
        raise CannotContinueException(
                """Cannot correct UMIs when kit is not Bioo.""")

    if reject_umi_errors and correct_umis:
        raise CannotContinueException(
                "Doesn't make sense to reject and *also* correct erroneous UMIs!!"
                " If passing --correct, you must also pass --keep-bad-umis.")

    # Which store to use
    if args['--store'] not in (STORE_OPTION_LMDB, STORE_OPTION_MEMORY):
        raise CannotContinueException("""Store {} is not supported.""".format(args['--store']))
    store = args['--store']

    outdir = args_to_out_dir(args)

    return (kit, store, outdir, input_file, paired, build_read_and_loc_dbs,
            reject_umi_errors, correct_umis, write_dedupped_sam,
            write_flagged_sam, write_dup_only_sam, write_dup_group_sam_like,
            write_umi_error_rejects, write_sam_headers, random_seed,
            debug_switch, dump_rg_db, dump_loc_db, dump_dup_group_db,
            dump_dup_db, dump_umi_error_db)

def run(kit, store, outdir, input_file, paired, build_read_and_loc_dbs,
        reject_umi_errors, correct_umis, write_dedupped_sam, write_flagged_sam,
        write_dup_only_sam, write_dup_group_sam_like, write_umi_error_rejects,
        write_sam_headers, random_seed, debug_switch, dump_rg_db, dump_loc_db,
        dump_dup_group_db, dump_dup_db, dump_umi_error_db):
    """Start the run.

    Args:
        kit (str): kit...
        store (str): Which storage backend to use.
        outdir (str): Output directory for results
        input_dir (str): ...
        build_read_and_loc_dbs (bool): Whether or not to build the
            read_group_db and location_db.
    """

    ## Set the random seed (if not set by user).
    if sys.version_info[0:2] in ((3,0), (3,1)):
        print("WARNING: You are using python v{}.{}, which has a random number "
                "generator which is not stable between runs.  Dupligänger "
                "output will not be the same between runs.  To fix this, switch "
                "to python version 2.7 or version >= 3.2.".format(
                    sys.version_info[0], sys.version_info[1]))
    random.seed(RANDOM_SEED if random_seed is None else random_seed)

    # report_db:
    #   A regular 'dict'.
    #   key: string describing a given (count) metric (I think they're all counts)
    #   val: (int) a count
    #   purpose: To provide a report at the end of the day on various metrics
    #       collected.
    # umi_error_db:
    #   A regular 'dict'.
    #   key: (str) A ReadGroup name.
    #   val: Not used (just doing 'True')
    #   purpose: To keep track of which reads have errors in one or more of
    #       their UMIs (and may optionally be rejected outright).
    # dup_group_db:
    #   A regular 'dict'.
    #   key: read_group_id
    #   val: list of read_group_ids that are duplicates of read_group_id
    #   purpose: To store stats, generate logs, debugging, etc. Maybe to
    #       also choose which duplicate to use.
    # dup_db:
    #   A SimpleBucketDB.
    #   key: ReadName of a PCR duplicate to be removed.
    #   val: Always just "True"
    #   purpose: When walking over the SAM/BAM file during last pass, any
    #       ReadName found in this DB will be removed from the final output.

    ## Setup databases
    report_db = setup_report_db()
    umi_error_db = {}
    dup_group_db = {}
    if store == STORE_OPTION_LMDB:
        db_file = os.path.join(outdir, os.path.split(input_file)[1] + '.sdd.db')
        parent_db = ParentDbLmdb(db_file, LMDB_MAX_DBS, LMDB_DB_SIZE)
        read_group_db = SimpleObjectDbLmdb('read_group', parent_db, parent_db.env, ReadGroup)
        location_bucket_store = SimpleBucketLmdb('location_bucket_store',
                parent_db, parent_db.env, DELIM_BUCKET_LIST, str)
        dup_db = SimpleBucketLmdb('duplicate', parent_db, parent_db.env,
                DELIM_BUCKET_LIST, str)
    elif store == STORE_OPTION_MEMORY:
        parent_db = ParentDbDict()
        read_group_db = SimpleObjectDbDict()
        location_bucket_store = SimpleBucketDict()
        dup_db = SimpleBucketDict()
    loc_db = LocationBucketDb(location_bucket_store, to_location_key_with_5p_trimming)

    ## What fp_process_location() and write_output_files() to use?
    if kit == KIT_BIOO:
        if paired:
            fp_process_location = functools.partial(
                    process_location_pe_bioo_1nt, report_db,
                    umi_error_db, reject_umi_errors, correct_umis)
            fp_write_output_files = write_output_files_pe
        else:
            raise CannotContinueException("""Doesn't support single-end yet.""")
            fp_process_location = functools.partial(
                    process_location_sr_bioo_1nt, report_db,
                    umi_error_db, reject_umi_errors, correct_umis)
            fp_write_output_files = write_output_files_sr
    else:
        if paired:
            fp_process_location = functools.partial(
                    process_location_pe_randomer_1nt, report_db)
            fp_write_output_files = write_output_files_pe
        else:
            raise CannotContinueException("""Doesn't support single-end yet.""")
            fp_process_location = functools.partial(
                    process_location_sr_randomer_1nt, report_db)
            fp_write_output_files = write_output_files_sr

    ## Winners and Losers
    # TODO: HARDCODED for now.
    fp_choose_winner_and_losers = choose_winner_and_losers_random_fixed_seed
    fp_log_winner_losers = log_winner_losers

    ## Setup output filenames
    # TODO: Offer BAM output?
    if write_dedupped_sam:
        dedupped_sam = filename_in_to_out_sambam(input_file,
                'dups_removed.sam', outdir)
    else:
        dedupped_sam = os.devnull
    if write_flagged_sam:
        flagged_sam = filename_in_to_out_sambam(input_file,
                'dups_flagged.sam', outdir)
    else:
        flagged_sam = os.devnull
    if write_dup_only_sam:
        dup_only_sam = filename_in_to_out_sambam(input_file,
                'duplicates.sam', outdir)
    else:
        dup_only_sam = os.devnull
    if write_dup_group_sam_like:
        dup_group_sam_like = filename_in_to_out_sambam(input_file,
                'dup_groups.samlike', outdir)
    else:
        dup_group_sam_like = os.devnull
    if write_umi_error_rejects:
        rejects_sam = filename_in_to_out_sambam(input_file,
                'umi_errors.sam', outdir)
    else:
        rejects_sam = os.devnull

    ### Go ###


    if build_read_and_loc_dbs:
        time1 = time.time()
        with sambamopen(input_file) as fin:

            ## First op, build read and location dbs
            write_to_read_and_location_dbs(report_db, fin, parent_db,
                    read_group_db, loc_db, RECORDS_PER_TXN)
        time2 = time.time()
        print("Building read_group_db and loc_db took: {}s, current mem (MBs): {}".format(
            time2 - time1, memory_info(True)))

    ## Second op, build DupGroupDB
    time1 = time.time()
    write_to_dup_group_db(report_db, parent_db, read_group_db, loc_db,
            dup_group_db, RECORDS_PER_TXN, fp_process_location)
    time2 = time.time()
    print("Building dup_group_db took: {}s, current mem (MBs): {}".format(
        time2 - time1, memory_info(True)))

    # Third op, resolve duplicates and build DupDB
    time1 = time.time()
    random.seed(RANDOM_SEED)
    write_to_dup_db(report_db, parent_db, read_group_db, dup_group_db, dup_db,
            RECORDS_PER_TXN, fp_choose_winner_and_losers,
            fp_log_winner_losers)
    time2 = time.time()
    print("Building dup_db took: {}s, current mem (MBs): {}".format(
        time2 - time1, memory_info(True)))

    ## Setup tmp filenames (now that most of the work is complete)
    out_files = [dedupped_sam, flagged_sam, dup_only_sam, dup_group_sam_like,
            rejects_sam]
    (tmp_dedupped_sam, tmp_flagged_sam, tmp_dup_only_sam,
            tmp_dup_group_sam_like, tmp_rejects_sam) = tmpf_start(*out_files)

    ## Fourth op: Write the dup_group_sam_like file
    time1 = time.time()
    write_dup_group_sam_like_file(parent_db, read_group_db, dup_group_db,
            tmp_dup_group_sam_like)
    time2 = time.time()
    print("Writing DupGroup file took: {}s, current mem (MBs): {}".format(
        time2 - time1, memory_info(True)))

    ## Fifth op: Walk through SAM/BAM input_file and write output files.
    time1 = time.time()
    fp_write_output_files(parent_db, read_group_db, dup_db, umi_error_db,
            input_file, reject_umi_errors, tmp_dedupped_sam, tmp_flagged_sam,
            tmp_dup_only_sam, tmp_rejects_sam, write_dedupped_sam,
            write_flagged_sam, write_dup_only_sam, write_dup_group_sam_like,
            write_umi_error_rejects, write_sam_headers)
    time2 = time.time()
    print("Writing output files took: {}s, current mem (MBs): {}".format(
        time2 - time1, memory_info(True)))

    # Write report_db
    for k,v in sorted(iteritems(report_db)):
        print("{}: {}".format(k,v))

    # Finish temp files
    tmpf_finish(tmp_dedupped_sam, tmp_flagged_sam, tmp_dup_only_sam,
            tmp_dup_group_sam_like, tmp_rejects_sam)

    # TEMPORARY!  Just for bootstrapping the testing of dedup. Will change to
    if dump_rg_db:
        sys.stderr.write(str(read_group_db))
        sys.stderr.write('\n')
    if dump_loc_db:
        sys.stderr.write(str(loc_db))
        sys.stderr.write('\n')
    if dump_dup_group_db:
        sys.stderr.write(str(dup_group_db))
        sys.stderr.write('\n')
    if dump_dup_db:
        sys.stderr.write(str(dup_db))
        sys.stderr.write('\n')
    if dump_umi_error_db:
        sys.stderr.write(str(umi_error_db))
        sys.stderr.write('\n')

    return out_files


###############
### Classes ###
###############

class DupGroup(set):
    """When completely built, represents a set of read_groups at one
    genomic location (technically at one location_key) that share the same UMI
    or UMIs.

    For example, for paired-end sequencing, to place two read-pairs in the same
    DupGroup, R1 and R2 from each read-pair would need to have the same
    synthetic_start (e.g. both read pairs have location_key
    'chr5:1000000:+,chr5:1000500:-') (or, in the future, possibly extend this)
    as well as share the same umi-pair (e.g. both R1s have ACGTACGT and both
    R2s have umi TTGGCCAA).

    Yup, DupGroup is just a set.
    """
    pass


############
### Main ###
############

def main():
    args = docopt(__doc__)
    run(*parse_args(args))

# vim: softtabstop=4:shiftwidth=4:expandtab



# -*- coding: utf-8 -*-
# Copyright (C) 2016, 2017, 2018  Jason Sydes and Peter Batzel
#
# This file is part of Dupligänger.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the License with this program.
#
# Written by Jason Sydes
# Conceptual Design by Peter Batzel and Jason Sydes

# Python 2/3 compatibility imports
from __future__ import absolute_import, division, print_function

# NOTE: Do *not* do the following:
# from builtins import str, chr, object
# py-lmdb uses bytes() for py3 and str() for py2.
# This package has different code for py2 and py3.
# And importing that future 'object' has a bug that screws up __slots__ in
# py2 (causes different behavior than in py3).

# Dupligänger imports
from dupliganger.constants import *
from dupliganger.exceptions import *

# Other imports

# For CustomList
import collections

#################
### Functions ###
#################

def to_location_key_no_5p_trim(read_group):
    """Convert a read_group to a location_key that looks like
    'chr5:12345:+,chr5:21111:+'.

    Basically, a double 1-nt entry, representing the soft-clipped-corrected
    start.

    Args:
        read_group (ReadGroup): Create a location key from read_group.
    """

    locs = []
    for read in read_group:
        syn_start, _, _, _ = parse_cigar(read.pos, read.strand, read.cigar)
        locs.append("{}:{}:{}".format(read.rname, syn_start, read.strand))
    return DELIM_BUCKET_LIST.join(locs)

def to_location_key_with_5p_trimming(read_group):
    """Convert a read_group to a location_key that looks like
    'chr5:1000000:+,chr5:1000500:-'.

    Basically, a double 1-nt entry, representing the soft-clipped-corrected
    and 5'-trimmed-corrected start.

    Args:
        read_group (ReadGroup): Create a location key from read_group.
    """
    # read qnames (and hence 5p trims, etc) are all identical w/in read_group

    # example: D12345:123:C7NMDANXX:6:1101:1184:39633-GGCCTAAT^AGCTCTAG;1^2
    trims_5p = [ int(trim) for trim in
            read_group[0].qname.split(DELIM_ANNO_TYPE)[1].split(
                DELIM_ANNO_READ_PAIR) ]

    locs = []
    for i, read in enumerate(read_group):
        syn_start, _, _, _ = parse_cigar(read.pos, read.strand, read.cigar)
        if read.strand == '+':
            syn_start_with_5p_trim = syn_start - trims_5p[i%2]
        else:
            syn_start_with_5p_trim = syn_start + trims_5p[i%2]
        locs.append("{}:{}:{}".format(read.rname, syn_start_with_5p_trim, read.strand))
    return DELIM_BUCKET_LIST.join(locs)

def parse_cigar(left, strand, cigar):
    """Parse the CIGAR string, and return the real and synthetic starts/ends,
    where 'synthetic' means that you add on a bit of NTs if there is
    soft-clipping.

    Args:
        left (int): SAM file 'POS' (leftmost alignment)
        strand (str): Forward or reverse strand ('+' or '-').
        cigar (str): CIGAR string.
    Returns:
        (int, int, int, int): 4-tuple: (synthetic_start, start, end,
            synthetic_end), where start/end are beginning and end of alignment
            respectively, and syn_start/syn_end are start/end corrected for
            soft-clipping.
    """
    clipped_left = None
    clipped_right = 0
    num = ''
    align_len = 0

    for c in cigar:
        if c in 'H':
            raise HardClippingNotSupportedException(
                    "Dupligänger does not support hard-clipping. "
                    "cigar: {}, left pos: {}, strand: {}".format(cigar, left,
                        strand))
        elif c in 'MIDNSP=X':
            # An operation
            if c == 'S':
                if clipped_left is None:
                    # clipping on left hand side
                    clipped_left = int(num)
                else:
                    # clipping on right hand side
                    clipped_right = int(num)
            elif c in 'M=XDN':
                align_len += int(num)
            elif c in 'IP':
                pass
            else:
                raise
            # Now reset number
            num = ''
            # Additionally, if clipped_left is still None at this point, then
            # the first operation was *not* a left-most soft clipping.  Set it
            # to False to indicate as much, so that if we encounter clipping on
            # the right hand side, it will fail the test 'clipped_left is
            # None', and right side clipping will be correctly recorded.
            if clipped_left is None:
                clipped_left = False
        else:
            # Another number
            num += c

    if clipped_left is None or clipped_left is False:
        clipped_left = 0

    if strand == '+':
        syn_start = left - clipped_left
        start = left
        end = start + align_len - 1
        syn_end = end + clipped_right
    else:
        syn_end = left - clipped_left
        end = left
        start = end + align_len - 1
        syn_start = start + clipped_right

    return (syn_start, start, end, syn_end)


###############
### Classes ###
###############

class Read(object):
    """An alignment line in a SAM file"""
    #_sam_fields = "qname flag rname pos mapq cigar rnext pnext tlen".split()
    _sam_fields = "qname flag rname pos mapq cigar".split()
    __slots__ = _sam_fields

    def __init__(self, line):
        p = line.split('\t')
        self.qname = p[0]
        self.flag  = int(p[1])
        self.rname = p[2]
        self.pos   = int(p[3])
        self.mapq  = p[4]
        self.cigar = p[5]
        # self.rnext = p[6]
        # self.pnext = p[7]
        # self.tlen  = p[8]

    def __repr__(self):
        """Note: This function is used to convert this object to string
        representation when storing entries in LMDB."""
        return DELIM_SAM_FIELD.join((self.qname, str(self.flag), self.rname,
            str(self.pos), self.mapq, self.cigar))

    @property
    def strand(self):
        return '-' if 0x010 == (0x010 & self.flag) else '+'


class ReadGroup(object):
    """Represents multiple alignment lines in a SAM file that all have the same
    read name.  (Really, this class should have been named AlignmentGroup.)"""
    __slots__ = ['_list']

    delim_list = DELIM_SAM_LIST_LMDB

    # ABC methods normally required by collections.MutableSequence
    def __delitem__(self, key):
        return self._list.__delitem__(key)
    def __getitem__(self, key):
        return self._list.__getitem__(key)
    def __setitem__(self, key, value):
        return self.__setitem__(key, value)
    def __len__(self):
        return self._list.__len__()
    def insert(self, index, obj):
        return self._list.insert(index, obj)

    # Our methods
    def __init__(self, l=[]):
        self._list = list(l)
    def __repr__(self):
        """Note: This function is used to convert this object to string
        representation when storing entries in LMDB."""
        return self._list.__repr__()
    def append(self, read):
        self._list.append(read)
    def load(self, obj_repr):
        """Note: This function is used to convert the string representation of
        this object in LMDB to an instance of this class.

        Args:
            obj_repr (str): In-database (e.g. LMDB) representation of the
                instance.

        Returns:
            ReadGroup: An instance of this class instantiated (a.k.a. loaded)
                from obj_repr.
        """
        self._list = [ Read(read_str) for read_str in
                obj_repr.split(self.__class__.delim_list) ]
        return self

    def __repr__(self):
        """Note: This function is used to convert this object to string
        representation when storing entries in LMDB."""
        return self.__class__.delim_list.join(
                [str(read) for read in self._list])

    @property
    def name(self):
        """Just returns the qname of the first Read in this ReadGroup."""
        return self[0].qname

    @property
    def reads(self):
        return self._list


# vim: softtabstop=4:shiftwidth=4:expandtab
